;;; qmath-mode.el --- Major mode for editing QMath files

;; Copyright (C) 1992,1993,1994  Tim Peters
;; Copyright (C) 2001,2002,2003 Alberto Gonz'alez Palomo

;; Derived from: python-mode.el by
;;         1995-1998 Barry A. Warsaw
;;         1992-1994 Tim Peters
;; Author: Alberto Gonz'alez Palomo
;; Maintainer: matmota@matracas.org
;; Created:    May 2001
;; Modified:   2004-02-05 Alberto Gonz'alez Palomo
;; Modified:   2007-01-30 Alberto Gonz'alez Palomo
;; Keywords:   qmath openmath omdoc

(defconst qmath-version "0.3.9"
  "'qmath-mode' version number.")

;; This software is provided as-is, without express or implied
;; warranty.  Permission to use, copy, modify, distribute or sell this
;; software, without fee, for any purpose and by any individual or
;; organization, is hereby granted, provided that the above copyright
;; notice and this paragraph appear in all copies.

;;; Commentary:

;; This is a major mode for editing QMath documents, derived from
;; python-mode.el, which was developed by Tim Peters after an original
;; idea by Michael A. Guravage.  Tim subsequently left the net; in 1995,
;; Barry Warsaw inherited the mode and is the current maintainer (of
;; python-mode.el). Tim's now back but disavows all responsibility for
;; the mode. Smart Tim :-)

;; This version of qmath-mode.el is no longer compatible with Emacs
;; 18.  I am striving to maintain compatibility with the X/Emacs 19
;; lineage but as time goes on that becomes more and more difficult.
;; I current recommend that you upgrade to the latest stable released
;; version of your favorite branch: Emacs 20.3 or better, or XEmacs
;; 20.4 or better (XEmacs 21.0 is in beta testing as of this writing
;; 27-Oct-1998 appears to work fine with this version of
;; qmath-mode.el).  Even Windows users should be using at least
;; NTEmacs 20.3, and XEmacs 21.0 will work very nicely on Windows when
;; it is released.

;; FOR MORE INFORMATION:

;; For more information on installing qmath-mode.el, especially with
;; respect to compatibility information, please see
;;
;;     http://www.matracas.org/qmath/qmath-mode/
;;

;; BUG REPORTING:

;; To submit bug reports, use C-c C-b.  Please include a complete, but
;; concise code sample and a recipe for reproducing the bug.  Send
;; suggestions and other comments to qmath-mode@matracas.org.

;; When in a QMath mode buffer, do a C-h m for more help.  It's
;; doubtful that a texinfo manual would be very useful, but if you
;; want to contribute one, I'll certainly accept it!

;; TO DO LIST:

;; - Better integration with pdb.py and gud-mode for debugging.
;; - Rewrite according to GNU Emacs Lisp standards.
;; - have qmath-execute-region on indented code act as if the region is
;;   left justified.  Avoids syntax errors.
;; - add a qmath-goto-block-down, bound to C-c C-d

;;; Code:

(require 'comint)
(require 'custom)
(eval-when-compile
  (require 'cl)
  (if (not (and (condition-case nil
		    (require 'custom)
		  (error nil))
		;; Stock Emacs 19.34 has a broken/old Custom library
		;; that does more harm than good.  Fortunately, it is
		;; missing defcustom
		(fboundp 'defcustom)))
      (error "STOP! STOP! STOP! STOP!

The Custom library was not found or is out of date.  A more current
version is required.  Please download and install the latest version
of the Custom library from:

    <http://www.dina.kvl.dk/~abraham/custom/>

See the QMath Mode home page for details:

    <http://www.matracas.org/qmath/qmath-mode>
")))



;; user definable variables
;; vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv

(defgroup qmath nil
  "Support for the QMath document language, <http://www.matracas.org/qmath>"
  :group 'languages
  :prefix "qmath-")

(defcustom qmath-qmath-command "QMath"
  "*Shell command used to start QMath interpreter."
  :type 'string
  :group 'qmath)

(defcustom qmath-jqmath-command "jqmath"
  "*Shell command used to start the JQMath interpreter."
  :type 'string
  :group 'qmath
  :tag "JQMath Command")

(defcustom qmath-default-interpreter 'cqmath
  "*Which QMath interpreter is used by default.
The value for this variable can be either `cqmath' or `jqmath'.

When the value is `cqmath', the variables `qmath-qmath-command' and
`qmath-qmath-command-args' are consulted to determine the interpreter
and arguments to use.

When the value is `jqmath', the variables `qmath-jqmath-command' and
`qmath-jqmath-command-args' are consulted to determine the interpreter
and arguments to use.

Note that this variable is consulted only the first time that a QMath
mode buffer is visited during an Emacs session.  After that, use
\\[qmath-toggle-shells] to change the interpreter shell."
  :type '(choice (const :tag "QMath (a.k.a. CQMath)" cqmath)
		 (const :tag "JQMath" jqmath))
  :group 'qmath)

(defcustom qmath-qmath-command-args '("-i")
  "*List of string arguments to be used when starting a QMath shell."
  :type '(repeat string)
  :group 'qmath)

(defcustom qmath-jqmath-command-args '("-i")
  "*List of string arguments to be used when starting a JQMath shell."
  :type '(repeat string)
  :group 'qmath
  :tag "JQMath Command Args")

(defcustom qmath-indent-offset 0
  "*Amount of offset per level of indentation.
`\\[qmath-guess-indent-offset]' can usually guess a good value when
you're editing someone else's QMath code."
  :type 'integer
  :group 'qmath)

(defcustom qmath-smart-indentation nil
  "*Should `qmath-mode' try to automagically set some indentation variables?
When this variable is non-nil, two things happen when a buffer is set
to `qmath-mode':

    1. `qmath-indent-offset' is guessed from existing code in the buffer.
       Only guessed values between 2 and 8 are considered.  If a valid
       guess can't be made (perhaps because you are visiting a new
       file), then the value in `qmath-indent-offset' is used.

    2. `indent-tabs-mode' is turned off if `qmath-indent-offset' does not
       equal `tab-width' (`indent-tabs-mode' is never turned on by
       QMath mode).  This means that for newly written code, tabs are
       only inserted in indentation if one tab is one indentation
       level, otherwise only spaces are used.

Note that both these settings occur *after* `qmath-mode-hook' is run,
so if you want to defeat the automagic configuration, you must also
set `qmath-smart-indentation' to nil in your `qmath-mode-hook'."
  :type 'boolean
  :group 'qmath)

(defcustom qmath-align-multiline-strings-p nil
  "*Flag describing how multi-line triple quoted strings are aligned.
When this flag is non-nil, continuation lines are lined up under the
preceding line's indentation.  When this flag is nil, continuation
lines are aligned to column zero."
  :type '(choice (const :tag "Align under preceding line" t)
		 (const :tag "Align to column zero" nil))
  :group 'qmath)

(defcustom qmath-block-comment-prefix "||| "
  "*String used by \\[comment-region] to comment out a block of code.
This should follow the convention for non-indenting comment lines so
that the indentation commands won't get confused (i.e., the string
should be of the form `#x...' where `x' is not a blank or a tab, and
`...' is arbitrary).  However, this string should not end in whitespace."
  :type 'string
  :group 'qmath)

(defcustom qmath-honor-comment-indentation nil
  "*Controls how comment lines influence subsequent indentation.

When nil, all comment lines are skipped for indentation purposes, and
if possible, a faster algorithm is used (i.e. X/Emacs 19 and beyond).

When t, lines that begin with a single `#' are a hint to subsequent
line indentation.  If the previous line is such a comment line (as
opposed to one that starts with `qmath-block-comment-prefix'), then its
indentation is used as a hint for this line's indentation.  Lines that
begin with `qmath-block-comment-prefix' are ignored for indentation
purposes.

When not nil or t, comment lines that begin with a `#' are used as
indentation hints, unless the comment character is in column zero."
  :type '(choice
	  (const :tag "Skip all comment lines (fast)" nil)
	  (const :tag "Single # `sets' indentation for next line" t)
	  (const :tag "Single # `sets' indentation except at column zero"
		 other)
	  )
  :group 'qmath)

(defcustom qmath-temp-directory
  (let ((ok '(lambda (x)
	       (and x
		    (setq x (expand-file-name x)) ; always true
		    (file-directory-p x)
		    (file-writable-p x)
		    x))))
    (or (funcall ok (getenv "TMPDIR"))
	(funcall ok "/usr/tmp")
	(funcall ok "/tmp")
	(funcall ok  ".")
	(error
	 "Couldn't find a usable temp directory -- set `qmath-temp-directory'")))
  "*Directory used for temp files created by a *QMath* process.
By default, the first directory from this list that exists and that you
can write into:  the value (if any) of the environment variable TMPDIR,
/usr/tmp, /tmp, or the current directory."
  :type 'string
  :group 'qmath)

(defcustom qmath-beep-if-tab-change t
  "*Ring the bell if `tab-width' is changed.
If a comment of the form

  \t# vi:set tabsize=<number>:

is found before the first code line when the file is entered, and the
current value of (the general Emacs variable) `tab-width' does not
equal <number>, `tab-width' is set to <number>, a message saying so is
displayed in the echo area, and if `qmath-beep-if-tab-change' is non-nil
the Emacs bell is also rung as a warning."
  :type 'boolean
  :group 'qmath)

(defcustom qmath-jump-on-exception t
  "*Jump to innermost exception frame in *QMath Output* buffer.
When this variable is non-nil and an exception occurs when running
QMath code synchronously in a subprocess, jump immediately to the
source code of the innermost traceback frame."
  :type 'boolean
  :group 'qmath)

(defcustom qmath-ask-about-save t
  "If not nil, ask about which buffers to save before executing some code.
Otherwise, all modified buffers are saved without asking."
  :type 'boolean
  :group 'qmath)

(defcustom qmath-backspace-function 'backward-delete-char-untabify
  "*Function called by `qmath-electric-backspace' when deleting backwards."
  :type 'function
  :group 'qmath)

(defcustom qmath-delete-function 'delete-char
  "*Function called by `qmath-electric-delete' when deleting forwards."
  :type 'function
  :group 'qmath)

(defcustom qmath-imenu-show-method-args-p nil 
  "*Controls echoing of arguments of functions & methods in the Imenu buffer.
When non-nil, arguments are printed."
  :type 'boolean
  :group 'qmath)
(make-variable-buffer-local 'qmath-indent-offset)

;; Document templates: set before calling the hook to allow customization.
(defcustom qmath-template-document-alist
  '(
    ("es" . '("QMATH 0.3.9\n[<-id_del_documento]\n\n:es\n\nContexto:\"Matemticas/OMDoc\"\n\n:\"Ttulo del documento\"\n:Nombre del autor\n:" (qmath-date) "\n"))
    ("en" . '("QMATH 0.3.9\n[<-document_id]\n\n:en\n\nContext:\"Mathematics/OMDoc\"\n\n:\"Document title\"\n:Author name\n:" (qmath-date) "\n"))
    ("de" . '("QMATH 0.3.9\n[<-dokument_id]\n\n:de\n\nContext:\"Mathematik/OMDoc\"\n\n:\"Titel der Dokuments\"\n:Name der Autors\n:" (qmath-date) "\n"))
    
    )
  "*Templates to insert when opening a new file."
  :type 'alist
  :group 'qmath)

;(qmath-append-document-template
; "es" '("QMATH 0.3.7\n[<-id_del_documento]\n\n:\"Ttulo del documento\"\n:Nombre del autor\n:" (qmath-date) "\n:es\n\nContexto:\"Matemticas/OMDoc\"\n\n")
;)
;(qmath-append-document-template
; "en" '("QMATH 0.3.7\n[<-document_id]\n\n:\"Document title\"\n:Author name\n:" (qmath-date) "\n:en\n\nContext:\"Mathematics/OMDoc\"\n\n")
; )
;(qmath-append-document-template
; "de" '("QMATH 0.3.7\n[<-dokument_id]\n\n:\"Titel der Dokuments\"\n:Name der Autors\n:" (qmath-date) "\n:de\n\nContext:\"Mathematik/OMDoc\"\n\n")
; )

;; Not customizable
(defvar qmath-master-file nil
  "If non-nil, execute the named file instead of the buffer's file.
The intent is to allow you to set this variable in the file's local
variable section, e.g.:

    # Local Variables:
    # qmath-master-file: \"master.qmath\"
    # End:

so that typing \\[qmath-execute-buffer] in that buffer executes the named
master file instead of the buffer's file.  If the file name has a
relative path, the value of variable `default-directory' for the
buffer is prepended to come up with a file name.")
(make-variable-buffer-local 'qmath-master-file)



;; ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
;; NO USER DEFINABLE VARIABLES BEYOND THIS POINT

(defconst qmath-emacs-features
  (let (features)
   ;; NTEmacs 19.34.6 has a broken make-temp-name; it always returns
   ;; the same string.
   (let ((tmp1 (make-temp-name ""))
	 (tmp2 (make-temp-name "")))
     (if (string-equal tmp1 tmp2)
	 (push 'broken-temp-names features)))
   ;; return the features
   features)
  "A list of features extant in the Emacs you are using.
There are many flavors of Emacs out there, with different levels of
support for features needed by `qmath-mode'.")

(defun qmath-date ()
  "Return current date formatted for insertion in a metadata entry."
  (replace-regexp-in-string
   "\\([+-][0-9]\\{2\\}\\)\\([0-9]\\{2\\}\\)" "\\1:\\2"
   (format-time-string "%Y-%m-%d %H:%M:%S%z")
   t nil)
  )

(defun qmath-insert-date ()
  "Insert current date at point, in QMath format."
  (interactive)
  (eval (list 'insert '(qmath-date)))
)

(defun qmath-user-language-tag ()
  "Return language tag, extracted from: LANG, LC_ALL, or LC_MESSAGES.\nIf not defined, return nil."
  (if (stringp (getenv "LANG"))
      (getenv "LANG")
    (if (stringp (getenv "LC_ALL"))
	(getenv "LC_ALL")
      (if (stringp (getenv "LC_MESSAGES"))
	  (getenv "LC_MESSAGES")
	(progn
	  (message "No language environment variable set. Use LANG, LC_ALL, or LC_MESSAGES.")
	  nil
	  )
	)
      )
    )
  )

(defun qmath-append-document-template (language-tag template)
  "Append a document template for the given language to
   qmath-template-document-alist."
   (interactive "sLanguage tag:\nMTemplate:")
   (setq qmath-template-document-alist
         (cons '(language-tag . template) qmath-template-document-alist)
	 )
   )

(defun qmath-template-document (language-tag)
  "Find document template for language, or nil."
  (cdr (assoc language-tag qmath-template-document-alist))
)

(defun qmath-insert-template-document (language-tag)
  "Insert a template document at the beginning of the buffer.
   First it tries with the full language tag, e.g. \"es_ES.UTF-8\".
   If there is no variable named \"document-template-es_ES.UTF-8\",
   it tries with \"es_ES\", and then with \"es\"."
  (interactive (list (completing-read "Language: "
				 qmath-template-document-alist
				 nil t nil nil nil nil
				 )
		     )
	       )
  (if (not language-tag)
      (setq language-tag (completing-read "Language: "
					  qmath-template-document-alist
					  nil t nil nil nil nil
					  )
	    )
    )
  (if (equal language-tag "")
      (message "No language specified.")
    (let* ((document-template nil) (parts nil))
      (setq document-template (qmath-template-document language-tag))
      (if document-template
	  (eval (cons 'insert (eval document-template)))
	(progn
	  (setq parts (split-string language-tag "[.]"))
	  (setq document-template (qmath-template-document (car parts)))
	  (if document-template
	      (eval (cons 'insert (eval document-template)))
	    (progn
	      (setq parts (split-string language-tag "[_]"))
	      (setq document-template (qmath-template-document (car parts)))
	      (if document-template
		  (eval (cons 'insert (eval document-template)))
		(message "There is no template available for language '%s'.\nYou can set yours with (qmath-append-document-template LANGUAGE TEMPLATE)." language-tag)
		)
	      )
	    )
	  )
	)
      )
    )
  )

(defvar qmath-font-lock-keywords
  (let ((kw1 (mapconcat 'identity
			'("PARAGRAPH_TYPE" "PARAGRAPH_TYPE_LEVEL_[0-9]+"
			  "CONTAINER_TYPE" "CONTAINER_TYPE_LEVEL_[0-9]+"
			  "PARAGRAPH_SECTION_MARKER"
			  "METADATA"
			  "QMATH_DIRECTIVE"
			  "APPLICATION" "ATTRIBUTION" "BINDING"
			  "SYMBOL" "VARIABLE"
			  "OP_IMPL" "OP_OR" "OP_AND" "OP_EQ"
			  "OP_PLUS" "OP_PROD" "OP_EXP" "OP_FACT"
			  "OP_BIND"
			  )
			"\\|"))
	)
    (list
     ;; math
     '("\\(^\\|[^$]\\)\\$\\([^$]\\(\\(\\$\\$\\)*[^$]\\)*\\)\\$[^$]"
       2 'qmath-math-expression-face)
     ;; erroneous literal dollar signs
     '("[^$]\\(\\(\\$\\$\\)+\\$\\)[^$]"
       1 'font-lock-warning-face)
     ;; comments
     '("\\(^|| .*\\)"
       1 font-lock-comment-face)
     ;; keywords
     (cons (concat "\\b\\(" kw1 "\\)\\b") 1)
     ;(cons "\\(\\$\\)" 1); It was excessive to mark both exp. and delimiters.
     ;; paragraph begin
     (cons "^[ \t]*\\([^ \t\n{}:\"$]+\\):" 1)
     ;; paragraph end
     (cons "^[ \t]*\\(}[^ \t\n{}:\"$]+\\)" 1)
     ;; inline tag begin
     (cons "\\[\\([[:alpha:][:digit:]_][[:alpha:][:digit:]_]+>\\)" 1)
     ;; inline tag end
     (cons "\\(<\\)\\]" 1)
     ;; Block begin
     (cons "^\\({\\)$" 1)
     ;; Block end
     (cons "^\\(}\\)$" 1)
     (cons "^\\(QMATH\\) [0-9.]+$" 1)
     ;; metadata language tag
     '("^:\\(aa\\|ab\\|af\\|am\\|ar\\|as\\|ay\\|az\\|ba\\|be\\|bg\\|bh\\|bi\\|bn\\|bo\\|br\\|ca\\|co\\|cs\\|cy\\|da\\|de\\|dz\\|el\\|en\\|eo\\|es\\|et\\|eu\\|fa\\|fi\\|fj\\|fo\\|fr\\|fy\\|ga\\|gd\\|gl\\|gn\\|gu\\|gv\\|ha\\|he\\|hi\\|hr\\|hu\\|hy\\|ia\\|id\\|ie\\|ik\\|is\\|it\\|iu\\|ja\\|jw\\|ka\\|kk\\|kl\\|km\\|kn\\|ko\\|ks\\|ku\\|kw\\|ky\\|la\\|lb\\|ln\\|lo\\|lt\\|lv\\|mg\\|mi\\|mk\\|ml\\|mn\\|mo\\|mr\\|ms\\|mt\\|my\\|na\\|ne\\|nl\\|no\\|oc\\|om\\|or\\|pa\\|pl\\|ps\\|pt\\|qu\\|rm\\|rn\\|ro\\|ru\\|rw\\|sa\\|sd\\|se\\|sg\\|sh\\|si\\|sk\\|sl\\|sm\\|sn\\|so\\|sq\\|sr\\|ss\\|st\\|su\\|sv\\|sw\\|ta\\|te\\|tg\\|th\\|ti\\|tk\\|tl\\|tn\\|to\\|tr\\|ts\\|tt\\|tw\\|ug\\|uk\\|ur\\|uz\\|vi\\|vo\\|wo\\|xh\\|yi\\|yo\\|za\\|zh\\|zu\\)$"
       1 'qmath-lang-tag-face)
     ;; metadata title
     '("^:\"\\([^\n\"]+\\)\"$"
       1 'qmath-title-1-face)
     ;; metadata value
     '("^:\\([^ \t\n:\"$]*\\):"
       1 'font-lock-variable-name-face)
     ;; metadata date
     '("^:\\([0-9]\\{4,\\}-[0-9]\\{1,2\\}-[0-9]\\{1,2\\}\\( [0-9]\\{1,2\\}:[0-9]\\{1,2\\}:[0-9]\\{1,2\\}\\([+-][0-9]\\{2\\}:[0-9]\\{2\\}\\)?\\)\\)$"
       1 'qmath-date-face)
     ;; metadata author
     '("^:\\([^\"\n:]+\\)$"
       1 'qmath-author-face)
     ;; string
     '("\\(\"[^\"\n]*\"\\|\"\\)"
       1 'font-lock-string-face keep)
     ;; reference identifier
     '("\\[<-\\([a-zA-Z0-9_-]+\\)\\]"
       1 'qmath-reference-id-face)
     ;; reference
     '("\\[\\([a-zA-Z0-9_:-]+\\)\\( [^ ].* \\)?\\]"
       1 'qmath-reference-face)
     ))
  "Additional expressions to highlight in QMath mode.")
(put 'qmath-mode 'font-lock-defaults '(qmath-font-lock-keywords))

;; have to bind qmath-file-queue before installing the kill-emacs-hook
(defvar qmath-file-queue nil
  "Queue of QMath temp files awaiting execution.
Currently-active file is at the head of the list.")


;; Constants

(defconst qmath-stringlit-re
  (concat
   ;; These fail if backslash-quote ends the string (not worth
   ;; fixing?).  They precede the short versions so that the first two
   ;; quotes don't look like an empty short string.
   ;;
   "\\[[a-zA-Z_0-9:]+\\]"
   "\\|"
   "\"\\([^\"\n\\]\\|\\\\.\\)*\""	; double-quoted
   )
  "Regular expression matching a QMath string literal.")

(defconst qmath-continued-re
  ;; This is tricky because a trailing backslash does not mean
  ;; continuation if it's in a comment
  (concat
   "\\(" "[^#\"\n\\]" "\\|" qmath-stringlit-re "\\)*"
   "\\\\$")
  "Regular expression matching QMath backslash continuation lines.")
  
(defconst qmath-blank-or-comment-re "[ \t]*\\($\\|#\\)"
  "Regular expression matching a blank or comment line.")

(defconst qmath-outdent-re
  (concat "\\(" (mapconcat 'identity
			   '("else:"
			     "except\\(\\s +.*\\)?:"
			     "finally:"
			     "elif\\s +.*:")
			   "\\|")
	  "\\)")
  "Regular expression matching statements to be dedented one level.")
  
(defconst qmath-block-closing-keywords-re
  "\\(\n\n\\)"
  "Regular expression matching keywords which typically close a block.")

(defconst qmath-no-outdent-re
  (concat
   "\\("
   (mapconcat 'identity
	      (list "try:"
		    "except\\(\\s +.*\\)?:"
		    "while\\s +.*:"
		    "for\\s +.*:"
		    "if\\s +.*:"
		    "elif\\s +.*:"
		    (concat qmath-block-closing-keywords-re "[ \t\n]")
		    )
	      "\\|")
	  "\\)")
  "Regular expression matching lines not to dedent after.")

(defconst qmath-defun-start-re
  "^\\([ \t]*\\)\\([a-zA-Z_0-9]+\\):\\(\\[[a-zA-Z_0-9<-]+\\]\\)?"
  ;; If you change this, you probably have to change qmath-current-defun
  ;; as well.  This is only used by qmath-current-defun to find the name
  ;; for add-log.el.
  "Regular expression matching a function, method, or variable assignment.")

(defconst qmath-class-start-re "^[ \t]*[a-zA-Z_0-9]+:\\(\\[[a-zA-Z_0-9<-]+\\]\\)?"
  ;; If you change this, you probably have to change qmath-current-defun
  ;; as well.  This is only used by qmath-current-defun to find the name
  ;; for add-log.el.
  "Regular expression for finding a class name.")

(defconst qmath-traceback-line-re
  "[ \t]+File \"\\([^\"]+\\)\", line \\([0-9]+\\)"
  "Regular expression that describes tracebacks.")



;; Major mode boilerplate

;; define a mode-specific abbrev table for those who use such things
(defvar qmath-mode-abbrev-table nil
  "Abbrev table in use in `qmath-mode' buffers.")
(define-abbrev-table 'qmath-mode-abbrev-table nil)

(defvar qmath-mode-hook nil
  "*Hook called by `qmath-mode'.")

(defvar qmath-mode-map ()
  "Keymap used in `qmath-mode' buffers.")
(if qmath-mode-map
    nil
  (setq qmath-mode-map (make-sparse-keymap))
  ;; electric keys
  (define-key qmath-mode-map ":" 'qmath-electric-colon)
  ;; indentation level modifiers
  (define-key qmath-mode-map "\C-c\C-l"  'qmath-shift-region-left)
  (define-key qmath-mode-map "\C-c\C-r"  'qmath-shift-region-right)
  (define-key qmath-mode-map "\C-c<"     'qmath-shift-region-left)
  (define-key qmath-mode-map "\C-c>"     'qmath-shift-region-right)
  ;; subprocess commands
  (define-key qmath-mode-map "\C-c\C-c"  'qmath-execute-buffer)
  (define-key qmath-mode-map "\C-c\C-m"  'qmath-execute-import-or-reload)
  (define-key qmath-mode-map "\C-c\C-s"  'qmath-execute-string)
  (define-key qmath-mode-map "\C-c|"     'qmath-execute-region)
  (define-key qmath-mode-map "\e\C-x"    'qmath-execute-def-or-class)
  (define-key qmath-mode-map "\C-c!"     'qmath-shell)
  (define-key qmath-mode-map "\C-c\C-t"  'qmath-toggle-shells)
  ;; Caution!  Enter here at your own risk.  We are trying to support
  ;; several behaviors and it gets disgusting. :-( This logic ripped
  ;; largely from CC Mode.
  ;;
  ;; In XEmacs 19, Emacs 19, and Emacs 20, we use this to bind
  ;; backwards deletion behavior to DEL, which both Delete and
  ;; Backspace get translated to.  There's no way to separate this
  ;; behavior in a clean way, so deal with it!  Besides, it's been
  ;; this way since the dawn of time.
  (if (not (boundp 'delete-key-deletes-forward))
      (define-key qmath-mode-map "\177" 'qmath-electric-backspace)
    ;; However, XEmacs 20 actually achieved enlightenment.  It is
    ;; possible to sanely define both backward and forward deletion
    ;; behavior under X separately (TTYs are forever beyond hope, but
    ;; who cares?  XEmacs 20 does the right thing with these too).
    (define-key qmath-mode-map [delete]    'qmath-electric-delete)
    (define-key qmath-mode-map [backspace] 'qmath-electric-backspace))
  ;; Separate M-BS from C-M-h.  The former should remain
  ;; backward-kill-word.
  (define-key qmath-mode-map [(control meta h)] 'qmath-mark-def-or-class)
  (define-key qmath-mode-map "\C-c\C-k"  'qmath-mark-block)
  ;; Miscellaneous
  (define-key qmath-mode-map "\C-c:"     'qmath-guess-indent-offset)
  (define-key qmath-mode-map "\C-c\t"    'qmath-indent-region)
  (define-key qmath-mode-map "\C-c\C-n"  'qmath-next-statement)
  (define-key qmath-mode-map "\C-c\C-p"  'qmath-previous-statement)
  (define-key qmath-mode-map "\C-c\C-u"  'qmath-goto-block-up)
  (define-key qmath-mode-map "\C-c#"     'qmath-comment-region)
  (define-key qmath-mode-map "\C-c?"     'qmath-describe-mode)
  (define-key qmath-mode-map "\C-c\C-hm" 'qmath-describe-mode)
  (define-key qmath-mode-map "\e\C-a"    'qmath-beginning-of-def-or-class)
  (define-key qmath-mode-map "\e\C-e"    'qmath-end-of-def-or-class)
  (define-key qmath-mode-map "\C-c-"     'qmath-up-exception)
  (define-key qmath-mode-map "\C-c="     'qmath-down-exception)
  ;; stuff that is `standard' but doesn't interface well with
  ;; qmath-mode, which forces us to rebind to special commands
  (define-key qmath-mode-map "\C-xnd"    'qmath-narrow-to-defun)
  ;; information
  (define-key qmath-mode-map "\C-c\C-b" 'qmath-submit-bug-report)
  (define-key qmath-mode-map "\C-c\C-v" 'qmath-version)
  ;; shadow global bindings for newline-and-indent w/ the qmath- version.
  ;; BAW - this is extremely bad form, but I'm not going to change it
  ;; for now.
  (mapcar #'(lambda (key)
	      (define-key qmath-mode-map key 'qmath-newline-and-indent))
	  (where-is-internal 'newline-and-indent))
  ;; Force RET to be qmath-newline-and-indent even if it didn't get
  ;; mapped by the above code.  motivation: Emacs' default binding for
  ;; RET is `newline' and C-j is `newline-and-indent'.  Most Qmatheers
  ;; expect RET to do a `qmath-newline-and-indent' and any Emacsers who
  ;; dislike this are probably knowledgeable enough to do a rebind.
  ;; However, we do *not* change C-j since many Emacsers have already
  ;; swapped RET and C-j and they don't want C-j bound to `newline' to 
  ;; change.
  (define-key qmath-mode-map "\C-m" 'qmath-newline-and-indent)
  ;; Deactivate tabulator indenting, since the indenting functions are
  ;; currently broken.
  (define-key qmath-mode-map [tab]   'dabbrev-expand); 'tab-to-tab-stop)
  )

(defvar qmath-mode-output-map nil
  "Keymap used in *QMath Output* buffers.")
(if qmath-mode-output-map
    nil
  (setq qmath-mode-output-map (make-sparse-keymap))
  (define-key qmath-mode-output-map [button2]  'qmath-mouseto-exception)
  (define-key qmath-mode-output-map "\C-c\C-c" 'qmath-goto-exception)
  ;; TBD: Disable all self-inserting keys.  This is bogus, we should
  ;; really implement this as *QMath Output* buffer being read-only
  (mapcar #' (lambda (key)
	       (define-key qmath-mode-output-map key
		 #'(lambda () (interactive) (beep))))
	     (where-is-internal 'self-insert-command))
  )

(defvar qmath-shell-map nil
  "Keymap used in *QMath* shell buffers.")
(if qmath-shell-map
    nil
  (setq qmath-shell-map (copy-keymap comint-mode-map))
  (define-key qmath-shell-map [tab]   'tab-to-tab-stop)
  (define-key qmath-shell-map "\C-c-" 'qmath-up-exception)
  (define-key qmath-shell-map "\C-c=" 'qmath-down-exception)
  )

(defvar qmath-mode-syntax-table nil
  "Syntax table used in `qmath-mode' buffers.")
(if qmath-mode-syntax-table
    nil
  (setq qmath-mode-syntax-table (make-syntax-table))
  (modify-syntax-entry ?\( "()" qmath-mode-syntax-table)
  (modify-syntax-entry ?\) ")(" qmath-mode-syntax-table)
  (modify-syntax-entry ?\[ "(]" qmath-mode-syntax-table)
  (modify-syntax-entry ?\] ")[" qmath-mode-syntax-table)
  (modify-syntax-entry ?\{ "(}" qmath-mode-syntax-table)
  (modify-syntax-entry ?\} "){" qmath-mode-syntax-table)
  ;; Add operator symbols misassigned in the std table
  (modify-syntax-entry ?\$ "$"  qmath-mode-syntax-table)
  (modify-syntax-entry ?\% "."  qmath-mode-syntax-table)
  (modify-syntax-entry ?\& "."  qmath-mode-syntax-table)
  (modify-syntax-entry ?\* "."  qmath-mode-syntax-table)
  (modify-syntax-entry ?\+ "."  qmath-mode-syntax-table)
  (modify-syntax-entry ?\- "."  qmath-mode-syntax-table)
  (modify-syntax-entry ?\/ "."  qmath-mode-syntax-table)
  (modify-syntax-entry ?\< "."  qmath-mode-syntax-table)
  (modify-syntax-entry ?\= "."  qmath-mode-syntax-table)
  (modify-syntax-entry ?\> "."  qmath-mode-syntax-table)
  (modify-syntax-entry ?\| "."  qmath-mode-syntax-table)
  ;; For historical reasons, underscore is word class instead of
  ;; symbol class.  GNU conventions say it should be symbol class, but
  ;; there's a natural conflict between what major mode authors want
  ;; and what users expect from `forward-word' and `backward-word'.
  ;; Guido and I have hashed this out and have decided to keep
  ;; underscore in word class.  If you're tempted to change it, try
  ;; binding M-f and M-b to qmath-forward-into-nomenclature and
  ;; qmath-backward-into-nomenclature instead.  This doesn't help in all
  ;; situations where you'd want the different behavior
  ;; (e.g. backward-kill-word).
  (modify-syntax-entry ?\_ "w"  qmath-mode-syntax-table)
  ;; Only double quote is string delimiter
  ;;(modify-syntax-entry ?\' "\"" qmath-mode-syntax-table)
  ;;(modify-syntax-entry ?\" "\"" qmath-mode-syntax-table)
  (modify-syntax-entry ?\" "." qmath-mode-syntax-table)
  ;; backquote is open and close paren
  (modify-syntax-entry ?\` "$"  qmath-mode-syntax-table)
  ;; comment delimiters
  ;;(modify-syntax-entry ?\# "<"  qmath-mode-syntax-table)
  ;;(modify-syntax-entry ?\n ">"  qmath-mode-syntax-table)
  )



;; Utilities

(defmacro qmath-safe (&rest body)
  "Safely execute BODY, return nil if an error occurred."
  (` (condition-case nil
	 (progn (,@ body))
       (error nil))))

(defsubst qmath-keep-region-active ()
  "Keep the region active in XEmacs."
  ;; Ignore byte-compiler warnings you might see.  Also note that
  ;; FSF's Emacs 19 does it differently; its policy doesn't require us
  ;; to take explicit action.
  (and (boundp 'zmacs-region-stays)
       (setq zmacs-region-stays t)))

(defsubst qmath-point (position)
  "Returns the value of point at certain commonly referenced POSITIONs.
POSITION can be one of the following symbols:

  bol  -- beginning of line
  eol  -- end of line
  bod  -- beginning of def or class
  eod  -- end of def or class
  bob  -- beginning of buffer
  eob  -- end of buffer
  boi  -- back to indentation
  bos  -- beginning of statement

This function does not modify point or mark."
  (let ((here (point)))
    (cond
     ((eq position 'bol) (beginning-of-line))
     ((eq position 'eol) (end-of-line))
     ((eq position 'bod) (qmath-beginning-of-def-or-class))
     ((eq position 'eod) (qmath-end-of-def-or-class))
     ;; Kind of funny, I know, but useful for qmath-up-exception.
     ((eq position 'bob) (beginning-of-buffer))
     ((eq position 'eob) (end-of-buffer))
     ((eq position 'boi) (back-to-indentation))
     ((eq position 'bos) (qmath-goto-initial-line))
     (t (error "Unknown buffer position requested: %s" position))
     )
    (prog1
	(point)
      (goto-char here))))

(defsubst qmath-highlight-line (from to file line)
  (cond
   ((fboundp 'make-extent)
    ;; XEmacs
    (let ((e (make-extent from to)))
      (set-extent-property e 'mouse-face 'highlight)
      (set-extent-property e 'qmath-exc-info (cons file line))
      (set-extent-property e 'keymap qmath-mode-output-map)))
   (t
    ;; Emacs -- Please port this!
    )
   ))

(defun qmath-in-literal (&optional lim)
  "Return non-nil if point is in a QMath literal (a comment or string).
Optional argument LIM indicates the beginning of the containing form,
i.e. the limit on how far back to scan."
  ;; This is the version used for non-XEmacs, which has a nicer
  ;; interface.
  ;;
  ;; WARNING: Watch out for infinite recursion.
  (let* ((lim (or lim (qmath-point 'bod)))
	 (state (parse-partial-sexp lim (point))))
    (cond
     ((nth 3 state) 'string)
     ((nth 4 state) 'comment)
     (t nil))))

;; XEmacs has a built-in function that should make this much quicker.
;; In this case, lim is ignored
(defun qmath-fast-in-literal (&optional lim)
  "Fast version of `qmath-in-literal', used only by XEmacs.
Optional LIM is ignored."
  ;; don't have to worry about context == 'block-comment
  (buffer-syntactic-context))

(if (fboundp 'buffer-syntactic-context)
    (defalias 'qmath-in-literal 'qmath-fast-in-literal))



;; Menu definitions, only relevent if you have the easymenu.el package
;; (standard in the latest Emacs 19 and XEmacs 19 distributions).
(defvar qmath-menu nil
  "Menu for QMath Mode.
This menu will get created automatically if you have the `easymenu'
package.  Note that the latest X/Emacs releases contain this package.")

(and (qmath-safe (require 'easymenu) t); Reactivated for now.
     (easy-menu-define
      qmath-menu qmath-mode-map "QMath Mode menu"
      '("QMath"
        ["Insert template document"  qmath-insert-template-document]
	["Comment Out Region"   qmath-comment-region  (mark)]
	["Uncomment Region"     (qmath-comment-region (point) (mark) '(4)) (mark)]
	"-"
	["Mark current block"   qmath-mark-block t]
	["Mark current def"     qmath-mark-def-or-class t]
	["Mark current class"   (qmath-mark-def-or-class t) t]
	"-"
	["Shift region left"    qmath-shift-region-left (mark)]
	["Shift region right"   qmath-shift-region-right (mark)]
	"-"
	["Import/reload file"   qmath-execute-import-or-reload t]
	["Execute buffer"       qmath-execute-buffer t]
	["Execute region"       qmath-execute-region (mark)]
	["Execute def or class" qmath-execute-def-or-class (mark)]
	["Execute string"       qmath-execute-string t]
	["Start interpreter..." qmath-shell t]
	"-"
	["Go to start of block" qmath-goto-block-up t]
	["Go to start of class" (qmath-beginning-of-def-or-class t) t]
	["Move to end of class" (qmath-end-of-def-or-class t) t]
	["Move to start of def" qmath-beginning-of-def-or-class t]
	["Move to end of def"   qmath-end-of-def-or-class t]
	"-"
	["Describe mode"        qmath-describe-mode t]
	)))



;; Imenu definitions
(defvar qmath-imenu-class-regexp
  (concat				; <<classes>>
   "\\("				;
   "\\[<-"				; We search for identifiers.
   "\\([a-zA-Z0-9_]+\\)"		; identifier name
   "\\(\\]\\)"				;
   "\\)"				; >>classes<<
   )
  "Regexp for QMath classes for use with the Imenu package."
  )

(defvar qmath-imenu-method-regexp
  (concat                               ; <<methods and functions>>
   "\\(\\["                             ; 
   "\\([a-zA-Z_0-9]+:"                  ;
   "\\([a-zA-Z_0-9]+\\)"                ;   name is here
   "\\)\\]"                             ;
   "\\)"                                ; >>methods and functions<<
;    "\\("                                ; 
;    "^[ \t]*"                            ; new line and maybe whitespace
;    "\\(def[ \t]+"                       ; function definitions start with def
;    "\\([a-zA-Z0-9_]+\\)"                ;   name is here
;                                         ;   function arguments...
; ;;   "[ \t]*(\\([-+/a-zA-Z0-9_=,\* \t\n.()\"'#]*\\))"
;    "[ \t]*(\\([^:#]*\\))"
;    "\\)"                                ; end of def
;    "[ \t]*:"                            ; and then the :
;    "\\)"                                ; >>methods and functions<<
   )
  "Regexp for QMath methods/functions for use with the Imenu package."
  )

(defvar qmath-imenu-method-no-arg-parens '(2 8)
  "Indices into groups of the QMath regexp for use with Imenu.

Using these values will result in smaller Imenu lists, as arguments to
functions are not listed.

See the variable `qmath-imenu-show-method-args-p' for more
information.")

(defvar qmath-imenu-method-arg-parens '(2 7)
  "Indices into groups of the QMath regexp for use with imenu.
Using these values will result in large Imenu lists, as arguments to
functions are listed.

See the variable `qmath-imenu-show-method-args-p' for more
information.")

;; Note that in this format, this variable can still be used with the
;; imenu--generic-function. Otherwise, there is no real reason to have
;; it.
(defvar qmath-imenu-generic-expression
  (cons
   (concat 
    qmath-imenu-class-regexp
;    "\\|"				; or...
;    qmath-imenu-method-regexp
    )
   qmath-imenu-method-no-arg-parens)
  "Generic QMath expression which may be used directly with Imenu.
Used by setting the variable `imenu-generic-expression' to this value.
Also, see the function \\[qmath-imenu-create-index] for a better
alternative for finding the index.")

;; These next two variables are used when searching for the QMath
;; class/definitions. Just saving some time in accessing the
;; generic-qmath-expression, really.
(defvar qmath-imenu-generic-regexp nil)
(defvar qmath-imenu-generic-parens nil)


(defun qmath-imenu-create-index-function ()
  "QMath interface function for the Imenu package.
Finds all QMath classes and functions/methods. Calls function
\\[qmath-imenu-create-index-engine].  See that function for the details
of how this works."
  (setq qmath-imenu-generic-regexp (car qmath-imenu-generic-expression)
	qmath-imenu-generic-parens (if qmath-imenu-show-method-args-p
				    qmath-imenu-method-arg-parens
				  qmath-imenu-method-no-arg-parens))
  (goto-char (point-min))
  ;; Warning: When the buffer has no classes or functions, this will
  ;; return nil, which seems proper according to the Imenu API, but
  ;; causes an error in the XEmacs port of Imenu.  Sigh.
  (qmath-imenu-create-index-engine nil))

(defun qmath-imenu-create-index-engine (&optional start-indent)
  "Function for finding Imenu definitions in QMath.

Finds all definitions (classes, methods, or functions) in a QMath
file for the Imenu package.

Returns a possibly nested alist of the form

	(INDEX-NAME . INDEX-POSITION)

The second element of the alist may be an alist, producing a nested
list as in

	(INDEX-NAME . INDEX-ALIST)

This function should not be called directly, as it calls itself
recursively and requires some setup.  Rather this is the engine for
the function \\[qmath-imenu-create-index-function].

It works recursively by looking for all definitions at the current
indention level.  When it finds one, it adds it to the alist.  If it
finds a definition at a greater indentation level, it removes the
previous definition from the alist. In its place it adds all
definitions found at the next indentation level.  When it finds a
definition that is less indented then the current level, it returns
the alist it has created thus far.

The optional argument START-INDENT indicates the starting indentation
at which to continue looking for QMath classes, methods, or
functions.  If this is not supplied, the function uses the indentation
of the first definition found."
  (let (index-alist
	sub-method-alist
	looking-p
	def-name prev-name
	cur-indent def-pos
	(class-paren (first  qmath-imenu-generic-parens)) 
	(def-paren   (second qmath-imenu-generic-parens)))
    (setq looking-p
	  (re-search-forward qmath-imenu-generic-regexp (point-max) t))
    (while looking-p
      (save-excursion
	;; used to set def-name to this value but generic-extract-name
	;; is new to imenu-1.14. this way it still works with
	;; imenu-1.11
	;;(imenu--generic-extract-name qmath-imenu-generic-parens))
	(let ((cur-paren (if (match-beginning class-paren)
			     class-paren def-paren)))
	  (setq def-name
		(buffer-substring-no-properties (match-beginning cur-paren)
						(match-end cur-paren))))
	(save-match-data
	  (qmath-beginning-of-def-or-class 'either))
	(beginning-of-line)
	(setq cur-indent (current-indentation)))
      ;; HACK: want to go to the next correct definition location.  We
      ;; explicitly list them here but it would be better to have them
      ;; in a list.
      (setq def-pos
	    (or (match-beginning class-paren)
		(match-beginning def-paren)))
      ;; if we don't have a starting indent level, take this one
      (or start-indent
	  (setq start-indent cur-indent))
      ;; if we don't have class name yet, take this one
      (or prev-name
	  (setq prev-name def-name))
      ;; what level is the next definition on?  must be same, deeper
      ;; or shallower indentation
      (cond
       ;; at the same indent level, add it to the list...
       ((= start-indent cur-indent)
	(push (cons def-name def-pos) index-alist))
       ;; deeper indented expression, recurse
       ((< start-indent cur-indent)
	;; the point is currently on the expression we're supposed to
	;; start on, so go back to the last expression. The recursive
	;; call will find this place again and add it to the correct
	;; list
	(re-search-backward qmath-imenu-generic-regexp (point-min) 'move)
	(setq sub-method-alist (qmath-imenu-create-index-engine cur-indent))
	(if sub-method-alist
	    ;; we put the last element on the index-alist on the start
	    ;; of the submethod alist so the user can still get to it.
	    (let ((save-elmt (pop index-alist)))
	      (push (cons prev-name
			  (cons save-elmt sub-method-alist))
		    index-alist))))
       ;; found less indented expression, we're done.
       (t 
	(setq looking-p nil)
	(re-search-backward qmath-imenu-generic-regexp (point-min) t)))
      ;; end-cond
      (setq prev-name def-name)
      (and looking-p
	   (setq looking-p
		 (re-search-forward qmath-imenu-generic-regexp
				    (point-max) 'move))))
    (nreverse index-alist)))


;;;###autoload
(defun qmath-mode ()
  "Major mode for editing QMath files.
To submit a problem report, enter `\\[qmath-submit-bug-report]' from a
`qmath-mode' buffer.  Do `\\[qmath-describe-mode]' for detailed
documentation.  To see what version of `qmath-mode' you are running,
enter `\\[qmath-version]'.

This mode knows about QMath indentation, tokens, comments and
continuation lines.  Paragraphs are separated by blank lines only.

COMMANDS
\\{qmath-mode-map}
VARIABLES

qmath-indent-offset\t\tindentation increment
qmath-block-comment-prefix\t\tcomment string used by `comment-region'
qmath-qmath-command\t\tshell command to invoke QMath interpreter
qmath-temp-directory\t\tdirectory used for temp files (if needed)
qmath-beep-if-tab-change\t\tring the bell if `tab-width' is changed"
  (interactive)
  ;; set up local variables
  (kill-all-local-variables)
  (make-local-variable 'font-lock-defaults)
  (make-local-variable 'paragraph-separate)
  (make-local-variable 'paragraph-start)
  (make-local-variable 'require-final-newline)
  (make-local-variable 'comment-start)
  (make-local-variable 'comment-end)
  (make-local-variable 'comment-start-skip)
  (make-local-variable 'comment-column)
  (make-local-variable 'comment-indent-function)
  (make-local-variable 'indent-region-function)
  (make-local-variable 'indent-line-function)
  (make-local-variable 'add-log-current-defun-function)
  ;;
  (set-syntax-table qmath-mode-syntax-table)
  (setq major-mode              'qmath-mode
	mode-name               "QMath"
	local-abbrev-table      qmath-mode-abbrev-table
	font-lock-defaults      '(qmath-font-lock-keywords)
	paragraph-separate      "^[ \t]*$"
	paragraph-start         "^[ \t]*$"
	require-final-newline   t
	comment-start           "|| "
	comment-end             ""
	comment-start-skip      "|| *"
	comment-column          40
	comment-indent-function 'qmath-comment-indent-function
	indent-region-function  'qmath-indent-region
	indent-line-function    'qmath-indent-line
	;; tell add-log.el how to find the current function/method/variable
	add-log-current-defun-function 'qmath-current-defun
	)
  (use-local-map qmath-mode-map)
  ;; add the menu
  (if qmath-menu
      (easy-menu-add qmath-menu))
  ;; Emacs 19 requires this
  (if (boundp 'comment-multi-line)
      (setq comment-multi-line nil))
  ;; Install Imenu if available
  (when (qmath-safe (require 'imenu))
    (setq imenu-create-index-function #'qmath-imenu-create-index-function)
    (setq imenu-generic-expression qmath-imenu-generic-expression)
    (if (fboundp 'imenu-add-to-menubar)
	(imenu-add-to-menubar (format "%s-%s" "IM" mode-name)))
    )
  ;; Create special faces
  (make-face 'qmath-title-1-face "Metadata: Title")
  (make-face 'qmath-title-2-face "Secondary title (not used yet)")
  (make-face 'qmath-author-face "Metadata: Author name")
  (make-face 'qmath-date-face "Metadata: Date")
  (make-face 'qmath-tabulation-face "Leading space for paragraphs (not used yet)")
  (make-face 'qmath-math-expression-face "Mathematical expressions (OMOBJs in OMDoc)")
  (make-face 'qmath-lang-tag-face "Metadata: Language tag ('es', 'de', 'en', etc...)")
  (make-face 'qmath-reference-id-face "Reference identifiers: [<identifier]")
  (make-face 'qmath-reference-face "References: [identifier]")
  ;; Default face properties
  (if (string-lessp emacs-version "21")
      (let* (this-frame (selected-frame))
	;; More modest settings for Emacs < 21
	(set-face-bold-p 'qmath-title-1-face t)
	(set-face-bold-p 'qmath-title-2-face t)
	(set-face-italic-p 'qmath-lang-tag-face t)
	(set-face-underline-p 'qmath-reference-face t)
	)
    (let* (this-frame (selected-frame))
      (set-face-attribute 'qmath-title-1-face this-frame
			  :family "helv" :weight 'bold :height 1.4)
      (set-face-attribute 'qmath-title-2-face this-frame
			  :height 1.2 :inherit 'qmath-title-1-face)
      (set-face-attribute 'qmath-author-face this-frame
			  :family "helv" :height 0.8)
      (set-face-attribute 'qmath-date-face this-frame
			  :family "helv" :height 0.8)
      (set-face-attribute 'qmath-tabulation-face this-frame
			  :family "fixed")
      (set-face-attribute 'qmath-math-expression-face this-frame
			  :family "symbol" :height 1.0 :weight 'normal
			  :slant 'normal
			  :underline nil :overline nil :strike-through nil
			  :box '(:line-width 1 ;:color "Gray40"
					     :style pressed-button))
      (set-face-attribute 'qmath-lang-tag-face this-frame
			  :family "fixed" :weight 'normal :slant 'italic)
      (set-face-attribute 'qmath-reference-id-face this-frame
			  :family "fixed" :weight 'normal :height 0.6)
      (set-face-attribute 'qmath-reference-face this-frame :underline t
			  :inherit 'qmath-reference-id-face)
      )
    )
  ;; dabbrev expansion doesn't change case.
  (setq dabbrev-case-fold-search nil)
  (setq dabbrev-case-replace     nil)
  ;; Run the mode hook.
  (if qmath-mode-hook
      (run-hooks 'qmath-mode-hook)
    (run-hooks 'qmath-mode-hook))
  ;; Now do the automagical guessing
  (if qmath-smart-indentation
    (let ((offset qmath-indent-offset))
      ;; It's okay if this fails to guess a good value
      (if (and (qmath-safe (qmath-guess-indent-offset))
	       (<= qmath-indent-offset 8)
	       (>= qmath-indent-offset 2))
	  (setq offset qmath-indent-offset))
      (setq qmath-indent-offset offset)
      ;; Only turn indent-tabs-mode off if tab-width !=
      ;; qmath-indent-offset.  Never turn it on, because the user must
      ;; have explicitly turned it off.
      (if (/= tab-width qmath-indent-offset)
	  (setq indent-tabs-mode nil))
      ))
  ;; Set the default shell if not already set
  ;(when (null qmath-which-shell)
  ;  (qmath-toggle-shells qmath-default-interpreter))
  
  ;; Write a document template if document is void.
  (if (equal (point-min) (point-max))
      (progn
	(qmath-insert-template-document (qmath-user-language-tag))
	(not-modified)
	)
    )
  )


;; electric characters
(defun qmath-outdent-p ()
  "Returns non-nil if the current line should dedent one level."
  (save-excursion
    (and (progn (back-to-indentation)
		(looking-at qmath-outdent-re))
	 ;; short circuit infloop on illegal construct
	 (not (bobp))
	 (progn (forward-line -1)
		(qmath-goto-initial-line)
		(back-to-indentation)
		(while (or (looking-at qmath-blank-or-comment-re)
			   (bobp))
		  (backward-to-indentation 1))
		(not (looking-at qmath-no-outdent-re)))
	 )))
      
(defun qmath-electric-colon (arg)
  "Insert a colon.
In certain cases the line is dedented appropriately.  If a numeric
argument ARG is provided, that many colons are inserted
non-electrically.  Electric behavior is inhibited inside a string or
comment."
  (interactive "P")
  (self-insert-command (prefix-numeric-value arg))
  ;; are we in a string or comment?
  (if (save-excursion
	(let ((pps (parse-partial-sexp (save-excursion
					 (qmath-beginning-of-def-or-class)
					 (point))
				       (point))))
	  (not (or (nth 3 pps) (nth 4 pps)))))
      (save-excursion
	(let ((here (point))
	      (outdent 0)
	      (indent (qmath-compute-indentation t)))
	  (if (and (not arg)
		   (qmath-outdent-p)
		   (= indent (save-excursion
			       (qmath-next-statement -1)
			       (qmath-compute-indentation t)))
		   )
	      (setq outdent qmath-indent-offset))
	  ;; Don't indent, only dedent.  This assumes that any lines
	  ;; that are already dedented relative to
	  ;; qmath-compute-indentation were put there on purpose.  It's
	  ;; highly annoying to have `:' indent for you.  Use TAB, C-c
	  ;; C-l or C-c C-r to adjust.  TBD: Is there a better way to
	  ;; determine this???
	  (if (< (current-indentation) indent) nil
	    (goto-char here)
	    (beginning-of-line)
	    (delete-horizontal-space)
	    (indent-to (- indent outdent))
	    )))))


;; QMath subprocess utilities and filters
(defun qmath-execute-file (proc filename)
  "Send to QMath interpreter process PROC \"execfile('FILENAME')\".
Make that process's buffer visible and force display.  Also make
comint believe the user typed this string so that
`kill-output-from-shell' does The Right Thing."
  (let ((curbuf (current-buffer))
	(procbuf (process-buffer proc))
;	(comint-scroll-to-bottom-on-output t)
	(msg (format "## working on region in file %s...\n" filename))
	(cmd (format "execfile(r'%s')\n" filename)))
    (unwind-protect
	(save-excursion
	  (set-buffer procbuf)
	  (goto-char (point-max))
	  (move-marker (process-mark proc) (point))
	  (funcall (process-filter proc) proc msg))
      (set-buffer curbuf))
    (process-send-string proc cmd)))

(defun qmath-comint-output-filter-function (string)
  "Watch output for QMath prompt and exec next file waiting in queue.
This function is appropriate for `comint-output-filter-functions'."
  ;; TBD: this should probably use split-string
  (when (and (or (string-equal string ">>> ")
		 (and (>= (length string) 5)
		      (string-equal (substring string -5) "\n>>> ")))
	     qmath-file-queue)
    (qmath-safe (delete-file (car qmath-file-queue)))
    (setq qmath-file-queue (cdr qmath-file-queue))
    (if qmath-file-queue
	(let ((qmathproc (get-buffer-process (current-buffer))))
	  (qmath-execute-file qmathproc (car qmath-file-queue))))
    ))

(defun qmath-postprocess-output-buffer (buf)
  "Highlight exceptions found in BUF.
If an exception occurred return t, otherwise return nil.  BUF must exist."
  (let (line file bol err-p)
    (save-excursion
      (set-buffer buf)
      (beginning-of-buffer)
      (while (re-search-forward qmath-traceback-line-re nil t)
	(setq file (match-string 1)
	      line (string-to-int (match-string 2))
	      bol (qmath-point 'bol))
	(qmath-highlight-line bol (qmath-point 'eol) file line)))
    (when (and qmath-jump-on-exception line)
      (beep)
      (qmath-jump-to-exception file line)
      (setq err-p t))
    err-p))



;;; Subprocess commands

;; only used when (memq 'broken-temp-names qmath-emacs-features)
(defvar qmath-serial-number 0)
(defvar qmath-exception-buffer nil)
(defconst qmath-output-buffer "*QMath Output*")
(make-variable-buffer-local 'qmath-output-buffer)

;; for toggling between CQMath and JQMath
(defvar qmath-which-shell nil)
(defvar qmath-which-args  qmath-qmath-command-args)
(defvar qmath-which-bufname "QMath")
(make-variable-buffer-local 'qmath-which-shell)
(make-variable-buffer-local 'qmath-which-args)
(make-variable-buffer-local 'qmath-which-bufname)

(defun qmath-toggle-shells (arg)
  "Toggles between the CQMath and JQMath shells.

With positive argument ARG (interactively \\[universal-argument]),
uses the CQMath shell, with negative ARG uses the JQMath shell, and
with a zero argument, toggles the shell.

Programmatically, ARG can also be one of the symbols `cqmath' or
`jqmath', equivalent to positive arg and negative arg respectively."
  (interactive "P")
  ;; default is to toggle
  (if (null arg)
      (setq arg 0))
  ;; preprocess arg
  (cond
   ((equal arg 0)
    ;; toggle
    (if (string-equal qmath-which-bufname "QMath")
	(setq arg -1)
      (setq arg 1)))
   ((equal arg 'cqmath) (setq arg 1))
   ((equal arg 'jqmath) (setq arg -1)))
  (let (msg)
    (cond
     ((< 0 arg)
      ;; set to CQMath
      (setq qmath-which-shell qmath-qmath-command
	    qmath-which-args qmath-qmath-command-args
	    qmath-which-bufname "QMath"
	    msg "CQMath"
	    mode-name "QMath"))
     ((> 0 arg)
      (setq qmath-which-shell qmath-jqmath-command
	    qmath-which-args qmath-jqmath-command-args
	    qmath-which-bufname "JQMath"
	    msg "JQMath"
	    mode-name "JQMath"))
     )
    (message "Using the %s shell" msg)
    (setq qmath-output-buffer (format "*%s Output*" qmath-which-bufname))))

;;;###autoload
(defun qmath-shell (&optional argprompt)
  "Start an interactive QMath interpreter in another window.
This is like Shell mode, except that QMath is running in the window
instead of a shell.  See the `Interactive Shell' and `Shell Mode'
sections of the Emacs manual for details, especially for the key
bindings active in the `*QMath*' buffer.

With optional \\[universal-argument], the user is prompted for the
flags to pass to the QMath interpreter.  This has no effect when this
command is used to switch to an existing process, only when a new
process is started.  If you use this, you will probably want to ensure
that the current arguments are retained (they will be included in the
prompt).  This argument is ignored when this function is called
programmatically, or when running in Emacs 19.34 or older.

Note: You can toggle between using the CQMath interpreter and the
JQMath interpreter by hitting \\[qmath-toggle-shells].  This toggles
buffer local variables which control whether all your subshell
interactions happen to the `*JQMath*' or `*QMath*' buffers (the
latter is the name used for the CQMath buffer).

Warning: Don't use an interactive QMath if you change sys.ps1 or
sys.ps2 from their default values, or if you're running code that
prints `>>> ' or `... ' at the start of a line.  `qmath-mode' can't
distinguish your output from QMath's output, and assumes that `>>> '
at the start of a line is a prompt from QMath.  Similarly, the Emacs
Shell mode code assumes that both `>>> ' and `... ' at the start of a
line are QMath prompts.  Bad things can happen if you fool either
mode.

Warning:  If you do any editing *in* the process buffer *while* the
buffer is accepting output from QMath, do NOT attempt to `undo' the
changes.  Some of the output (nowhere near the parts you changed!) may
be lost if you do.  This appears to be an Emacs bug, an unfortunate
interaction between undo and process filters; the same problem exists in
non-QMath process buffers using the default (Emacs-supplied) process
filter."
  (interactive "P")
  ;; Set the default shell if not already set
  (when (null qmath-which-shell)
    (qmath-toggle-shells qmath-default-interpreter))
  (let ((args qmath-which-args))
    (when (and argprompt
	       (interactive-p)
	       (fboundp 'split-string))
      ;; TBD: Perhaps force "-i" in the final list?
      (setq args (split-string
		  (read-string (concat qmath-which-bufname
				       " arguments: ")
			       (concat
				(mapconcat 'identity qmath-which-args " ") " ")
			       ))))
    (switch-to-buffer-other-window
     (apply 'make-comint qmath-which-bufname qmath-which-shell nil args))
    (make-local-variable 'comint-prompt-regexp)
    (setq comint-prompt-regexp "^>>> \\|^[.][.][.] \\|^(pdb) ")
    (add-hook 'comint-output-filter-functions
	      'qmath-comint-output-filter-function)
    (set-syntax-table qmath-mode-syntax-table)
    (use-local-map qmath-shell-map)
    ))

(defun qmath-clear-queue ()
  "Clear the queue of temporary files waiting to execute."
  (interactive)
  (let ((n (length qmath-file-queue)))
    (mapcar 'delete-file qmath-file-queue)
    (setq qmath-file-queue nil)
    (message "%d pending files de-queued." n)))


(defun qmath-execute-region (start end &optional async)
  "Execute the region in a QMath interpreter.

The region is first copied into a temporary file (in the directory
`qmath-temp-directory').  If there is no QMath interpreter shell
running, this file is executed synchronously using
`shell-command-on-region'.  If the program is long running, use
\\[universal-argument] to run the command asynchronously in its own
buffer.

When this function is used programmatically, arguments START and END
specify the region to execute, and optional third argument ASYNC, if
non-nil, specifies to run the command asynchronously in its own
buffer.

If the QMath interpreter shell is running, the region is execfile()'d
in that shell.  If you try to execute regions too quickly,
`qmath-mode' will queue them up and execute them one at a time when
it sees a `>>> ' prompt from QMath.  Each time this happens, the
process buffer is popped into a window (if it's not already in some
window) so you can see it, and a comment of the form

    \t## working on region in file <name>...

is inserted at the end.  See also the command `qmath-clear-queue'."
  (interactive "r\nP")
  (or (< start end)
      (error "Region is empty"))
  (let* ((proc (get-process qmath-which-bufname))
	 (temp (if (memq 'broken-temp-names qmath-emacs-features)
		   (let
		       ((sn qmath-serial-number)
			(pid (and (fboundp 'emacs-pid) (emacs-pid))))
		     (setq qmath-serial-number (1+ qmath-serial-number))
		     (if pid
			 (format "qmath-%d-%d" sn pid)
		       (format "qmath-%d" sn)))
		 (make-temp-name "qmath-")))
	 (file (expand-file-name temp qmath-temp-directory)))
    (write-region start end file nil 'nomsg)
    (cond
     ;; always run the code in its own asynchronous subprocess
     (async
      (let* ((buf (generate-new-buffer-name qmath-output-buffer))
	     ;; TBD: a horrible hack, but why create new Custom variables?
	     (arg (if (string-equal qmath-which-bufname "QMath")
		      "-u" "")))
	(start-process qmath-which-bufname buf qmath-which-shell arg file)
	(pop-to-buffer buf)
	(qmath-postprocess-output-buffer buf)
	))
     ;; if the QMath interpreter shell is running, queue it up for
     ;; execution there.
     (proc
      ;; use the existing qmath shell
      (if (not qmath-file-queue)
	  (qmath-execute-file proc file)
	(message "File %s queued for execution" file))
      (setq qmath-file-queue (append qmath-file-queue (list file)))
      (setq qmath-exception-buffer (cons file (current-buffer))))
     (t
      ;; TBD: a horrible hack, buy why create new Custom variables?
      (let ((cmd (concat qmath-which-shell
			 (if (string-equal qmath-which-bufname "JQMath")
			     " -" ""))))
	;; otherwise either run it synchronously in a subprocess
	(shell-command-on-region start end cmd qmath-output-buffer)
	;; shell-command-on-region kills the output buffer if it never
	;; existed and there's no output from the command
	(if (not (get-buffer qmath-output-buffer))
	    (message "No output.")
	  (setq qmath-exception-buffer (current-buffer))
	  (let ((err-p (qmath-postprocess-output-buffer qmath-output-buffer)))
	    (pop-to-buffer qmath-output-buffer)
	    (if err-p
		(pop-to-buffer qmath-exception-buffer)))
	  )))
     )))


;; Code execution commands
(defun qmath-execute-buffer (&optional async)
  "Send the contents of the buffer to a QMath interpreter.
If the file local variable `qmath-master-file' is non-nil, execute the
named file instead of the buffer's file.

If there is a *QMath* process buffer it is used.  If a clipping
restriction is in effect, only the accessible portion of the buffer is
sent.  A trailing newline will be supplied if needed.

See the `\\[qmath-execute-region]' docs for an account of some
subtleties, including the use of the optional ASYNC argument."
  (interactive "P")
  (if qmath-master-file
      (let* ((filename (expand-file-name qmath-master-file))
	     (buffer (or (get-file-buffer filename)
			 (find-file-noselect filename))))
	(set-buffer buffer)))
  (qmath-execute-region (point-min) (point-max) async))

(defun qmath-execute-import-or-reload (&optional async)
  "Import the current buffer's file in a QMath interpreter.

If the file has already been imported, then do reload instead to get
the latest version.

If the file's name does not end in \".qmath\", then do execfile instead.

If the current buffer is not visiting a file, do `qmath-execute-buffer'
instead.

If the file local variable `qmath-master-file' is non-nil, import or
reload the named file instead of the buffer's file.  The file may be
saved based on the value of `qmath-execute-import-or-reload-save-p'.

See the `\\[qmath-execute-region]' docs for an account of some
subtleties, including the use of the optional ASYNC argument.

This may be preferable to `\\[qmath-execute-buffer]' because:

 - Definitions stay in their module rather than appearing at top
   level, where they would clutter the global namespace and not affect
   uses of qualified names (MODULE.NAME).

 - The QMath debugger gets line number information about the functions."
  (interactive "P")
  ;; Check file local variable qmath-master-file
  (if qmath-master-file
      (let* ((filename (expand-file-name qmath-master-file))
             (buffer (or (get-file-buffer filename)
                         (find-file-noselect filename))))
        (set-buffer buffer)))
  (let ((file (buffer-file-name (current-buffer))))
    (if file
        (progn
	  ;; Maybe save some buffers
	  (save-some-buffers (not qmath-ask-about-save) nil)
          (qmath-execute-string
           (if (string-match "\\.qmath\\(\\.utf\\)?$" file)
               (let ((f (file-name-sans-extension
			 (file-name-nondirectory file))))
                 (format "if globals().has_key('%s'):\n    reload(%s)\nelse:\n    import %s\n"
                         f f f))
             (format "execfile(r'%s')\n" file))
           async))
      ;; else
      (qmath-execute-buffer async))))


(defun qmath-execute-def-or-class (&optional async)
  "Send the current function or class definition to a QMath interpreter.

If there is a *QMath* process buffer it is used.

See the `\\[qmath-execute-region]' docs for an account of some
subtleties, including the use of the optional ASYNC argument."
  (interactive "P")
  (save-excursion
    (qmath-mark-def-or-class)
    ;; mark is before point
    (qmath-execute-region (mark) (point) async)))


(defun qmath-execute-string (string &optional async)
  "Send the argument STRING to a QMath interpreter.

If there is a *QMath* process buffer it is used.

See the `\\[qmath-execute-region]' docs for an account of some
subtleties, including the use of the optional ASYNC argument."
  (interactive "sExecute QMath command: ")
  (save-excursion
    (set-buffer (get-buffer-create
                 (generate-new-buffer-name " *QMath Command*")))
    (insert string)
    (qmath-execute-region (point-min) (point-max) async)))



(defun qmath-jump-to-exception (file line)
  "Jump to the QMath code in FILE at LINE."
  (let ((buffer (cond ((string-equal file "<stdin>")
		       (if (consp qmath-exception-buffer)
			   (cdr qmath-exception-buffer)
			 qmath-exception-buffer))
		      ((and (consp qmath-exception-buffer)
			    (string-equal file (car qmath-exception-buffer)))
		       (cdr qmath-exception-buffer))
		      ((qmath-safe (find-file-noselect file)))
		      ;; could not figure out what file the exception
		      ;; is pointing to, so prompt for it
		      (t (find-file (read-file-name "Exception file: "
						    nil
						    file t))))))
    (pop-to-buffer buffer)
    ;; Force QMath mode
    (if (not (eq major-mode 'qmath-mode))
	(qmath-mode))
    (goto-line line)
    (message "Jumping to exception in file %s on line %d" file line)))

(defun qmath-mouseto-exception (event)
  "Jump to the code which caused the QMath exception at EVENT.
EVENT is usually a mouse click."
  (interactive "e")
  (cond
   ((fboundp 'event-point)
    ;; XEmacs
    (let* ((point (event-point event))
	   (buffer (event-buffer event))
	   (e (and point buffer (extent-at point buffer 'qmath-exc-info)))
	   (info (and e (extent-property e 'qmath-exc-info))))
      (message "Event point: %d, info: %s" point info)
      (and info
	   (qmath-jump-to-exception (car info) (cdr info)))
      ))
   ;; Emacs -- Please port this!
   ))

(defun qmath-goto-exception ()
  "Go to the line indicated by the traceback."
  (interactive)
  (let (file line)
    (save-excursion
      (beginning-of-line)
      (if (looking-at qmath-traceback-line-re)
	  (setq file (match-string 1)
		line (string-to-int (match-string 2)))))
    (if (not file)
	(error "Not on a traceback line"))
    (qmath-jump-to-exception file line)))

(defun qmath-find-next-exception (start buffer searchdir errwhere)
  "Find the next QMath exception and jump to the code that caused it.
START is the buffer position in BUFFER from which to begin searching
for an exception.  SEARCHDIR is a function, either
`re-search-backward' or `re-search-forward' indicating the direction
to search.  ERRWHERE is used in an error message if the limit (top or
bottom) of the trackback stack is encountered."
  (let (file line)
    (save-excursion
      (set-buffer buffer)
      (goto-char (qmath-point start))
      (if (funcall searchdir qmath-traceback-line-re nil t)
	  (setq file (match-string 1)
		line (string-to-int (match-string 2)))))
    (if (and file line)
	(qmath-jump-to-exception file line)
      (error "%s of traceback" errwhere))))

(defun qmath-down-exception (&optional bottom)
  "Go to the next line down in the traceback.
With \\[univeral-argument] (programmatically, optional argument
BOTTOM), jump to the bottom (innermost) exception in the exception
stack."
  (interactive "P")
  (let* ((proc (get-process "QMath"))
	 (buffer (if proc "*QMath*" qmath-output-buffer)))
    (if bottom
	(qmath-find-next-exception 'eob buffer 're-search-backward "Bottom")
      (qmath-find-next-exception 'eol buffer 're-search-forward "Bottom"))))

(defun qmath-up-exception (&optional top)
  "Go to the previous line up in the traceback.
With \\[universal-argument] (programmatically, optional argument TOP)
jump to the top (outermost) exception in the exception stack."
  (interactive "P")
  (let* ((proc (get-process "QMath"))
	 (buffer (if proc "*QMath*" qmath-output-buffer)))
    (if top
	(qmath-find-next-exception 'bob buffer 're-search-forward "Top")
      (qmath-find-next-exception 'bol buffer 're-search-backward "Top"))))


;; Electric deletion
(defun qmath-electric-backspace (arg)
  "Delete preceding character or levels of indentation.
Deletion is performed by calling the function in `qmath-backspace-function'
with a single argument (the number of characters to delete).

If point is at the leftmost column, delete the preceding newline.

Otherwise, if point is at the leftmost non-whitespace character of a
line that is neither a continuation line nor a non-indenting comment
line, or if point is at the end of a blank line, this command reduces
the indentation to match that of the line that opened the current
block of code.  The line that opened the block is displayed in the
echo area to help you keep track of where you are.  With
\\[universal-argument] dedents that many blocks (but not past column
zero).

Otherwise the preceding character is deleted, converting a tab to
spaces if needed so that only a single column position is deleted.
\\[universal-argument] specifies how many characters to delete;
default is 1.

When used programmatically, argument ARG specifies the number of
blocks to dedent, or the number of characters to delete, as indicated
above."
  (interactive "*p")
  (if (or (/= (current-indentation) (current-column))
	  (bolp)
	  (qmath-continuation-line-p)
;	  (not qmath-honor-comment-indentation)
;	  (looking-at "#[^ \t\n]")	; non-indenting #
	  )
      (funcall qmath-backspace-function arg)
    ;; else indent the same as the colon line that opened the block
    ;; force non-blank so qmath-goto-block-up doesn't ignore it
    (insert-char ?* 1)
    (backward-char)
    (let ((base-indent 0)		; indentation of base line
	  (base-text "")		; and text of base line
	  (base-found-p nil))
      (save-excursion
	(while (< 0 arg)
	  (condition-case nil		; in case no enclosing block
	      (progn
		(qmath-goto-block-up 'no-mark)
		(setq base-indent (current-indentation)
		      base-text   (qmath-suck-up-leading-text)
		      base-found-p t))
	    (error nil))
	  (setq arg (1- arg))))
      (delete-char 1)			; toss the dummy character
      (delete-horizontal-space)
      (indent-to base-indent)
      (if base-found-p
	  (message "Closes block: %s" base-text)))))


(defun qmath-electric-delete (arg)
  "Delete preceding or following character or levels of whitespace.

The behavior of this function depends on the variable
`delete-key-deletes-forward'.  If this variable is nil (or does not
exist, as in older Emacsen and non-XEmacs versions), then this
function behaves identically to \\[c-electric-backspace].

If `delete-key-deletes-forward' is non-nil and is supported in your
Emacs, then deletion occurs in the forward direction, by calling the
function in `qmath-delete-function'.

\\[universal-argument] (programmatically, argument ARG) specifies the
number of characters to delete (default is 1)."
  (interactive "*p")
  (if (or (and (fboundp 'delete-forward-p) ;XEmacs 21
	       (delete-forward-p))
	  (and (boundp 'delete-key-deletes-forward) ;XEmacs 20
	       delete-key-deletes-forward))
      (funcall qmath-delete-function arg)
    (qmath-electric-backspace arg)))

;; required for pending-del and delsel modes
(put 'qmath-electric-backspace 'delete-selection 'supersede) ;delsel
(put 'qmath-electric-backspace 'pending-delete   'supersede) ;pending-del
(put 'qmath-electric-delete    'delete-selection 'supersede) ;delsel
(put 'qmath-electric-delete    'pending-delete   'supersede) ;pending-del



(defun qmath-indent-line (&optional arg)
  "Fix the indentation of the current line according to QMath rules.
With \\[universal-argument] (programmatically, the optional argument
ARG non-nil), ignore dedenting rules for block closing statements
(e.g. return, raise, break, continue, pass)

This function is normally bound to `indent-line-function' so
\\[indent-for-tab-command] will call it."
  (interactive "P")
  (let* ((ci (current-indentation))
	 (move-to-indentation-p (<= (current-column) ci))
	 (need (qmath-compute-indentation (not arg))))
    ;; see if we need to dedent
    (if (qmath-outdent-p)
	(setq need (- need qmath-indent-offset)))
    (if (/= ci need)
	(save-excursion
	  (beginning-of-line)
	  (delete-horizontal-space)
	  (indent-to need)))
    (if move-to-indentation-p (back-to-indentation))))

(defun qmath-newline-and-indent ()
  "Strives to act like the Emacs `newline-and-indent'.
This is just `strives to' because correct indentation can't be computed
from scratch for QMath code.  In general, deletes the whitespace before
point, inserts a newline, and takes an educated guess as to how you want
the new line indented."
  (interactive)
  (let ((ci (current-indentation)))
    (if (< ci (current-column))		; if point beyond indentation
	(newline-and-indent)
      ;; else try to act like newline-and-indent "normally" acts
      (beginning-of-line)
      (insert-char ?\n 1)
      (move-to-column ci))))

(defun qmath-compute-indentation (honor-block-close-p)
  "Compute QMath indentation.
When HONOR-BLOCK-CLOSE-P is non-nil, statements such as `return',
`raise', `break', `continue', and `pass' force one level of
dedenting."
  (save-excursion
    (beginning-of-line)
    (let* ((bod (qmath-point 'bod))
	   (pps (parse-partial-sexp bod (point)))
	   (boipps (parse-partial-sexp bod (qmath-point 'boi)))
	   placeholder)
      (cond
       ;; are we inside a multi-line string or comment?
       ((or (and (nth 3 pps) (nth 3 boipps))
	    (and (nth 4 pps) (nth 4 boipps)))
	(save-excursion
	  (if (not qmath-align-multiline-strings-p) 0
	    ;; skip back over blank & non-indenting comment lines
	    ;; note: will skip a blank or non-indenting comment line
	    ;; that happens to be a continuation line too
	    (re-search-backward "^[ \t]*\\([^ \t\n#]\\|#[ \t\n]\\)" nil 'move)
	    (back-to-indentation)
	    (current-column))))
       ;; are we on a continuation line?
       ((qmath-continuation-line-p)
	(let ((startpos (point))
	      (open-bracket-pos (qmath-nesting-level))
	      endpos searching found state)
	  (if open-bracket-pos
	      (progn
		;; align with first item in list; else a normal
		;; indent beyond the line with the open bracket
		(goto-char (1+ open-bracket-pos)) ; just beyond bracket
		;; is the first list item on the same line?
		(skip-chars-forward " \t")
		(if (null (memq (following-char) '(?\n ?# ?\\)))
					; yes, so line up with it
		    (current-column)
		  ;; first list item on another line, or doesn't exist yet
		  (forward-line 1)
		  (while (and (< (point) startpos)
			      (looking-at "[ \t]*[#\n\\\\]")) ; skip noise
		    (forward-line 1))
		  (if (and (< (point) startpos)
			   (/= startpos
			       (save-excursion
				 (goto-char (1+ open-bracket-pos))
				 (forward-comment (point-max))
				 (point))))
		      ;; again mimic the first list item
		      (current-indentation)
		    ;; else they're about to enter the first item
		    (goto-char open-bracket-pos)
		    (setq placeholder (point))
		    (qmath-goto-initial-line)
		    (qmath-goto-beginning-of-tqs
		     (save-excursion (nth 3 (parse-partial-sexp
					     placeholder (point)))))
		    (+ (current-indentation) qmath-indent-offset))))

	    ;; else on backslash continuation line
	    (forward-line -1)
	    (if (qmath-continuation-line-p) ; on at least 3rd line in block
		(current-indentation)	; so just continue the pattern
	      ;; else started on 2nd line in block, so indent more.
	      ;; if base line is an assignment with a start on a RHS,
	      ;; indent to 2 beyond the leftmost "="; else skip first
	      ;; chunk of non-whitespace characters on base line, + 1 more
	      ;; column
	      (end-of-line)
	      (setq endpos (point)  searching t)
	      (back-to-indentation)
	      (setq startpos (point))
	      ;; look at all "=" from left to right, stopping at first
	      ;; one not nested in a list or string
	      (while searching
		(skip-chars-forward "^=" endpos)
		(if (= (point) endpos)
		    (setq searching nil)
		  (forward-char 1)
		  (setq state (parse-partial-sexp startpos (point)))
		  (if (and (zerop (car state)) ; not in a bracket
			   (null (nth 3 state))) ; & not in a string
		      (progn
			(setq searching nil) ; done searching in any case
			(setq found
			      (not (or
				    (eq (following-char) ?=)
				    (memq (char-after (- (point) 2))
					  '(?< ?> ?!)))))))))
	      (if (or (not found)	; not an assignment
		      (looking-at "[ \t]*\\\\")) ; <=><spaces><backslash>
		  (progn
		    (goto-char startpos)
		    (skip-chars-forward "^ \t\n")))
	      (1+ (current-column))))))

       ;; not on a continuation line
       ((bobp) (current-indentation))

       ;; Dfn: "Indenting comment line".  A line containing only a
       ;; comment, but which is treated like a statement for
       ;; indentation calculation purposes.  Such lines are only
       ;; treated specially by the mode; they are not treated
       ;; specially by the QMath interpreter.

       ;; The rules for indenting comment lines are a line where:
       ;;   - the first non-whitespace character is `#', and
       ;;   - the character following the `#' is whitespace, and
       ;;   - the line is dedented with respect to (i.e. to the left
       ;;     of) the indentation of the preceding non-blank line.

       ;; The first non-blank line following an indenting comment
       ;; line is given the same amount of indentation as the
       ;; indenting comment line.

       ;; All other comment-only lines are ignored for indentation
       ;; purposes.

       ;; Are we looking at a comment-only line which is *not* an
       ;; indenting comment line?  If so, we assume that it's been
       ;; placed at the desired indentation, so leave it alone.
       ;; Indenting comment lines are aligned as statements down
       ;; below.
       ((and (looking-at "[ \t]*#[^ \t\n]")
	     ;; NOTE: this test will not be performed in older Emacsen
	     (fboundp 'forward-comment)
	     (<= (current-indentation)
		 (save-excursion
		   (forward-comment (- (point-max)))
		   (current-indentation))))
	(current-indentation))

       ;; else indentation based on that of the statement that
       ;; precedes us; use the first line of that statement to
       ;; establish the base, in case the user forced a non-std
       ;; indentation for the continuation lines (if any)
       (t
	;; skip back over blank & non-indenting comment lines note:
	;; will skip a blank or non-indenting comment line that
	;; happens to be a continuation line too.  use fast Emacs 19
	;; function if it's there.
	(if (and (eq qmath-honor-comment-indentation nil)
		 (fboundp 'forward-comment))
	    (forward-comment (- (point-max)))
	  (let ((prefix-re (concat qmath-block-comment-prefix "[ \t]*"))
		done)
	    (while (not done)
	      (re-search-backward "^[ \t]*\\([^ \t\n#]\\|#\\)" nil 'move)
	      (setq done (or (bobp)
			     (and (eq qmath-honor-comment-indentation t)
				  (save-excursion
				    (back-to-indentation)
				    (not (looking-at prefix-re))
				    ))
			     (and (not (eq qmath-honor-comment-indentation t))
				  (save-excursion
				    (back-to-indentation)
				    (not (zerop (current-column)))))
			     ))
	      )))
	;; if we landed inside a string, go to the beginning of that
	;; string. this handles triple quoted, multi-line spanning
	;; strings.
	(qmath-goto-beginning-of-tqs (nth 3 (parse-partial-sexp bod (point))))
	;; now skip backward over continued lines
	(setq placeholder (point))
	(qmath-goto-initial-line)
	;; we may *now* have landed in a TQS, so find the beginning of
	;; this string.
	(qmath-goto-beginning-of-tqs
	 (save-excursion (nth 3 (parse-partial-sexp
				 placeholder (point)))))
	(+ (current-indentation)
	   (if (qmath-statement-opens-block-p)
	       qmath-indent-offset
	     (if (and honor-block-close-p (qmath-statement-closes-block-p))
		 (- qmath-indent-offset)
	       0)))
	)))))

(defun qmath-guess-indent-offset (&optional global)
  "Guess a good value for, and change, `qmath-indent-offset'.

By default, make a buffer-local copy of `qmath-indent-offset' with the
new value, so that other QMath buffers are not affected.  With
\\[universal-argument] (programmatically, optional argument GLOBAL),
change the global value of `qmath-indent-offset'.  This affects all
QMath buffers (that don't have their own buffer-local copy), both
those currently existing and those created later in the Emacs session.

Some people use a different value for `qmath-indent-offset' than you use.
There's no excuse for such foolishness, but sometimes you have to deal
with their ugly code anyway.  This function examines the file and sets
`qmath-indent-offset' to what it thinks it was when they created the
mess.

Specifically, it searches forward from the statement containing point,
looking for a line that opens a block of code.  `qmath-indent-offset' is
set to the difference in indentation between that line and the QMath
statement following it.  If the search doesn't succeed going forward,
it's tried again going backward."
  (interactive "P")			; raw prefix arg
  (let (new-value
	(start (point))
	(restart (point))
	(found nil)
	colon-indent)
    (qmath-goto-initial-line)
    (while (not (or found (eobp)))
      (when (and (re-search-forward ":[ \t]*\\($\\|[#\\]\\)" nil 'move)
		 (not (qmath-in-literal restart)))
	(setq restart (point))
	(qmath-goto-initial-line)
	(if (qmath-statement-opens-block-p)
	    (setq found t)
	  (goto-char restart))))
    (unless found
      (goto-char start)
      (qmath-goto-initial-line)
      (while (not (or found (bobp)))
	(setq found (and
		     (re-search-backward ":[ \t]*\\($\\|[#\\]\\)" nil 'move)
		     (or (qmath-goto-initial-line) t) ; always true -- side effect
		     (qmath-statement-opens-block-p)))))
    (setq colon-indent (current-indentation)
	  found (and found (zerop (qmath-next-statement 1)))
	  new-value (- (current-indentation) colon-indent))
    (goto-char start)
    (if (not found)
	(error "Sorry, couldn't guess a value for qmath-indent-offset")
      (funcall (if global 'kill-local-variable 'make-local-variable)
	       'qmath-indent-offset)
      (setq qmath-indent-offset new-value)
      (or noninteractive
	  (message "%s value of qmath-indent-offset set to %d"
		   (if global "Global" "Local")
		   qmath-indent-offset)))
    ))

(defun qmath-comment-indent-function ()
  "QMath version of `comment-indent-function'."
  ;; This is required when filladapt is turned off.  Without it, when
  ;; filladapt is not used, comments which start in column zero
  ;; cascade one character to the right
  (save-excursion
    (beginning-of-line)
    (let ((eol (qmath-point 'eol)))
      (and comment-start-skip
	   (re-search-forward comment-start-skip eol t)
	   (setq eol (match-beginning 0)))
      (goto-char eol)
      (skip-chars-backward " \t")
      (max comment-column (+ (current-column) (if (bolp) 0 1)))
      )))

(defun qmath-narrow-to-defun (&optional class)
  "Make text outside current defun invisible.
The defun visible is the one that contains point or follows point.
Optional CLASS is passed directly to `qmath-beginning-of-def-or-class'."
  (interactive "P")
  (save-excursion
    (widen)
    (qmath-end-of-def-or-class class)
    (let ((end (point)))
      (qmath-beginning-of-def-or-class class)
      (narrow-to-region (point) end))))


(defun qmath-shift-region (start end count)
  "Indent lines from START to END by COUNT spaces."
  (save-excursion
    (goto-char end)
    (beginning-of-line)
    (setq end (point))
    (goto-char start)
    (beginning-of-line)
    (setq start (point))
    (indent-rigidly start end count)))

(defun qmath-shift-region-left (start end &optional count)
  "Shift region of QMath code to the left.
The lines from the line containing the start of the current region up
to (but not including) the line containing the end of the region are
shifted to the left, by `qmath-indent-offset' columns.

If a prefix argument is given, the region is instead shifted by that
many columns.  With no active region, dedent only the current line.
You cannot dedent the region if any line is already at column zero."
  (interactive
   (let ((p (point))
	 (m (mark))
	 (arg current-prefix-arg))
     (if m
	 (list (min p m) (max p m) arg)
       (list p (save-excursion (forward-line 1) (point)) arg))))
  ;; if any line is at column zero, don't shift the region
  (save-excursion
    (goto-char start)
    (while (< (point) end)
      (back-to-indentation)
      (if (and (zerop (current-column))
	       (not (looking-at "\\s *$")))
	  (error "Region is at left edge"))
      (forward-line 1)))
  (qmath-shift-region start end (- (prefix-numeric-value
				 (or count qmath-indent-offset))))
  (qmath-keep-region-active))

(defun qmath-shift-region-right (start end &optional count)
  "Shift region of QMath code to the right.
The lines from the line containing the start of the current region up
to (but not including) the line containing the end of the region are
shifted to the right, by `qmath-indent-offset' columns.

If a prefix argument is given, the region is instead shifted by that
many columns.  With no active region, indent only the current line."
  (interactive
   (let ((p (point))
	 (m (mark))
	 (arg current-prefix-arg))
     (if m
	 (list (min p m) (max p m) arg)
       (list p (save-excursion (forward-line 1) (point)) arg))))
  (qmath-shift-region start end (prefix-numeric-value
			      (or count qmath-indent-offset)))
  (qmath-keep-region-active))

(defun qmath-indent-region (start end &optional indent-offset)
  "Reindent a region of QMath code.

The lines from the line containing the start of the current region up
to (but not including) the line containing the end of the region are
reindented.  If the first line of the region has a non-whitespace
character in the first column, the first line is left alone and the
rest of the region is reindented with respect to it.  Else the entire
region is reindented with respect to the (closest code or indenting
comment) statement immediately preceding the region.

This is useful when code blocks are moved or yanked, when enclosing
control structures are introduced or removed, or to reformat code
using a new value for the indentation offset.

If a numeric prefix argument is given, it will be used as the value of
the indentation offset.  Else the value of `qmath-indent-offset' will be
used.

Warning: The region must be consistently indented before this function
is called!  This function does not compute proper indentation from
scratch (that's impossible in QMath), it merely adjusts the existing
indentation to be correct in context.

Warning: This function really has no idea what to do with
non-indenting comment lines, and shifts them as if they were indenting
comment lines.  Fixing this appears to require telepathy.

Special cases: whitespace is deleted from blank lines; continuation
lines are shifted by the same amount their initial line was shifted,
in order to preserve their relative indentation with respect to their
initial line; and comment lines beginning in column 1 are ignored."
  (interactive "*r\nP")			; region; raw prefix arg
  (save-excursion
    (goto-char end)   (beginning-of-line) (setq end (point-marker))
    (goto-char start) (beginning-of-line)
    (let ((qmath-indent-offset (prefix-numeric-value
			     (or indent-offset qmath-indent-offset)))
	  (indents '(-1))		; stack of active indent levels
	  (target-column 0)		; column to which to indent
	  (base-shifted-by 0)		; amount last base line was shifted
	  (indent-base (if (looking-at "[ \t\n]")
			   (qmath-compute-indentation t)
			 0))
	  ci)
      (while (< (point) end)
	(setq ci (current-indentation))
	;; figure out appropriate target column
	(cond
	 ((or (eq (following-char) ?#)	; comment in column 1
	      (looking-at "[ \t]*$"))	; entirely blank
	  (setq target-column 0))
	 ((qmath-continuation-line-p)	; shift relative to base line
	  (setq target-column (+ ci base-shifted-by)))
	 (t				; new base line
	  (if (> ci (car indents))	; going deeper; push it
	      (setq indents (cons ci indents))
	    ;; else we should have seen this indent before
	    (setq indents (memq ci indents)) ; pop deeper indents
	    (if (null indents)
		(error "Bad indentation in region, at line %d"
		       (save-restriction
			 (widen)
			 (1+ (count-lines 1 (point)))))))
	  (setq target-column (+ indent-base
				 (* qmath-indent-offset
				    (- (length indents) 2))))
	  (setq base-shifted-by (- target-column ci))))
	;; shift as needed
	(if (/= ci target-column)
	    (progn
	      (delete-horizontal-space)
	      (indent-to target-column)))
	(forward-line 1))))
  (set-marker end nil))

(defun qmath-comment-region (beg end &optional arg)
  "Like `comment-region' but uses double hash (`#') comment starter."
  (interactive "r\nP")
  (let ((comment-start qmath-block-comment-prefix))
    (comment-region beg end arg)))


;; Functions for moving point
(defun qmath-previous-statement (count)
  "Go to the start of the COUNTth preceding QMath statement.
By default, goes to the previous statement.  If there is no such
statement, goes to the first statement.  Return count of statements
left to move.  `Statements' do not include blank, comment, or
continuation lines."
  (interactive "p")			; numeric prefix arg
  (if (< count 0) (qmath-next-statement (- count))
    (qmath-goto-initial-line)
    (let (start)
      (while (and
	      (setq start (point))	; always true -- side effect
	      (> count 0)
	      (zerop (forward-line -1))
	      (qmath-goto-statement-at-or-above))
	(setq count (1- count)))
      (if (> count 0) (goto-char start)))
    count))

(defun qmath-next-statement (count)
  "Go to the start of next QMath statement.
If the statement at point is the i'th QMath statement, goes to the
start of statement i+COUNT.  If there is no such statement, goes to the
last statement.  Returns count of statements left to move.  `Statements'
do not include blank, comment, or continuation lines."
  (interactive "p")			; numeric prefix arg
  (if (< count 0) (qmath-previous-statement (- count))
    (beginning-of-line)
    (let (start)
      (while (and
	      (setq start (point))	; always true -- side effect
	      (> count 0)
	      (qmath-goto-statement-below))
	(setq count (1- count)))
      (if (> count 0) (goto-char start)))
    count))

(defun qmath-goto-block-up (&optional nomark)
  "Move up to start of current block.
Go to the statement that starts the smallest enclosing block; roughly
speaking, this will be the closest preceding statement that ends with a
colon and is indented less than the statement you started on.  If
successful, also sets the mark to the starting point.

`\\[qmath-mark-block]' can be used afterward to mark the whole code
block, if desired.

If called from a program, the mark will not be set if optional argument
NOMARK is not nil."
  (interactive)
  (let ((start (point))
	(found nil)
	initial-indent)
    (qmath-goto-initial-line)
    ;; if on blank or non-indenting comment line, use the preceding stmt
    (if (looking-at "[ \t]*\\($\\|#[^ \t\n]\\)")
	(progn
	  (qmath-goto-statement-at-or-above)
	  (setq found (qmath-statement-opens-block-p))))
    ;; search back for colon line indented less
    (setq initial-indent (current-indentation))
    (if (zerop initial-indent)
	;; force fast exit
	(goto-char (point-min)))
    (while (not (or found (bobp)))
      (setq found
	    (and
	     (re-search-backward ":[ \t]*\\($\\|[#\\]\\)" nil 'move)
	     (or (qmath-goto-initial-line) t) ; always true -- side effect
	     (< (current-indentation) initial-indent)
	     (qmath-statement-opens-block-p))))
    (if found
	(progn
	  (or nomark (push-mark start))
	  (back-to-indentation))
      (goto-char start)
      (error "Enclosing block not found"))))

(defun qmath-beginning-of-def-or-class (&optional class count)
  "Move point to start of `def' or `class'.

Searches back for the closest preceding `def'.  If you supply a prefix
arg, looks for a `class' instead.  The docs below assume the `def'
case; just substitute `class' for `def' for the other case.
Programmatically, if CLASS is `either', then moves to either `class'
or `def'.

When second optional argument is given programmatically, move to the
COUNTth start of `def'.

If point is in a `def' statement already, and after the `d', simply
moves point to the start of the statement.

Otherwise (i.e. when point is not in a `def' statement, or at or
before the `d' of a `def' statement), searches for the closest
preceding `def' statement, and leaves point at its start.  If no such
statement can be found, leaves point at the start of the buffer.

Returns t iff a `def' statement is found by these rules.

Note that doing this command repeatedly will take you closer to the
start of the buffer each time.

To mark the current `def', see `\\[qmath-mark-def-or-class]'."
  (interactive "P")			; raw prefix arg
  (setq count (or count 1))
  (let ((at-or-before-p (<= (current-column) (current-indentation)))
	(start-of-line (goto-char (qmath-point 'bol)))
	(start-of-stmt (goto-char (qmath-point 'bos)))
	(start-re (cond ((eq class 'either) "^[ \t]*\\(class\\|def\\)\\>")
			(class "^[ \t]*class\\>")
			(t "^[ \t]*def\\>")))
	)
    ;; searching backward
    (if (and (< 0 count)
	     (or (/= start-of-stmt start-of-line)
		 (not at-or-before-p)))
	(end-of-line))
    ;; search forward
    (if (and (> 0 count)
	     (zerop (current-column))
	     (looking-at start-re))
	(end-of-line))
    (if (re-search-backward start-re nil 'move count)
	(goto-char (match-beginning 0)))))

;; Backwards compatibility
(defalias 'beginning-of-qmath-def-or-class 'qmath-beginning-of-def-or-class)

(defun qmath-end-of-def-or-class (&optional class count)
  "Move point beyond end of `def' or `class' body.

By default, looks for an appropriate `def'.  If you supply a prefix
arg, looks for a `class' instead.  The docs below assume the `def'
case; just substitute `class' for `def' for the other case.
Programmatically, if CLASS is `either', then moves to either `class'
or `def'.

When second optional argument is given programmatically, move to the
COUNTth end of `def'.

If point is in a `def' statement already, this is the `def' we use.

Else, if the `def' found by `\\[qmath-beginning-of-def-or-class]'
contains the statement you started on, that's the `def' we use.

Otherwise, we search forward for the closest following `def', and use that.

If a `def' can be found by these rules, point is moved to the start of
the line immediately following the `def' block, and the position of the
start of the `def' is returned.

Else point is moved to the end of the buffer, and nil is returned.

Note that doing this command repeatedly will take you closer to the
end of the buffer each time.

To mark the current `def', see `\\[qmath-mark-def-or-class]'."
  (interactive "P")			; raw prefix arg
  (if (and count (/= count 1))
      (qmath-beginning-of-def-or-class (- 1 count)))
  (let ((start (progn (qmath-goto-initial-line) (point)))
	(which (cond ((eq class 'either) "\\(class\\|def\\)")
		     (class "class")
		     (t "def")))
	(state 'not-found))
    ;; move point to start of appropriate def/class
    (if (looking-at (concat "[ \t]*" which "\\>")) ; already on one
	(setq state 'at-beginning)
      ;; else see if qmath-beginning-of-def-or-class hits container
      (if (and (qmath-beginning-of-def-or-class class)
	       (progn (qmath-goto-beyond-block)
		      (> (point) start)))
	  (setq state 'at-end)
	;; else search forward
	(goto-char start)
	(if (re-search-forward (concat "^[ \t]*" which "\\>") nil 'move)
	    (progn (setq state 'at-beginning)
		   (beginning-of-line)))))
    (cond
     ((eq state 'at-beginning) (qmath-goto-beyond-block) t)
     ((eq state 'at-end) t)
     ((eq state 'not-found) nil)
     (t (error "Internal error in `qmath-end-of-def-or-class'")))))

;; Backwards compabitility
(defalias 'end-of-qmath-def-or-class 'qmath-end-of-def-or-class)


;; Functions for marking regions
(defun qmath-mark-block (&optional extend just-move)
  "Mark following block of lines.  With prefix arg, mark structure.
Easier to use than explain.  It sets the region to an `interesting'
block of succeeding lines.  If point is on a blank line, it goes down to
the next non-blank line.  That will be the start of the region.  The end
of the region depends on the kind of line at the start:

 - If a comment, the region will include all succeeding comment lines up
   to (but not including) the next non-comment line (if any).

 - Else if a prefix arg is given, and the line begins one of these
   structures:

     if elif else try except finally for while def class

   the region will be set to the body of the structure, including
   following blocks that `belong' to it, but excluding trailing blank
   and comment lines.  E.g., if on a `try' statement, the `try' block
   and all (if any) of the following `except' and `finally' blocks
   that belong to the `try' structure will be in the region.  Ditto
   for if/elif/else, for/else and while/else structures, and (a bit
   degenerate, since they're always one-block structures) def and
   class blocks.

 - Else if no prefix argument is given, and the line begins a QMath
   block (see list above), and the block is not a `one-liner' (i.e.,
   the statement ends with a colon, not with code), the region will
   include all succeeding lines up to (but not including) the next
   code statement (if any) that's indented no more than the starting
   line, except that trailing blank and comment lines are excluded.
   E.g., if the starting line begins a multi-statement `def'
   structure, the region will be set to the full function definition,
   but without any trailing `noise' lines.

 - Else the region will include all succeeding lines up to (but not
   including) the next blank line, or code or indenting-comment line
   indented strictly less than the starting line.  Trailing indenting
   comment lines are included in this case, but not trailing blank
   lines.

A msg identifying the location of the mark is displayed in the echo
area; or do `\\[exchange-point-and-mark]' to flip down to the end.

If called from a program, optional argument EXTEND plays the role of
the prefix arg, and if optional argument JUST-MOVE is not nil, just
moves to the end of the block (& does not set mark or display a msg)."
  (interactive "P")			; raw prefix arg
  (qmath-goto-initial-line)
  ;; skip over blank lines
  (while (and
	  (looking-at "[ \t]*$")	; while blank line
	  (not (eobp)))			; & somewhere to go
    (forward-line 1))
  (if (eobp)
      (error "Hit end of buffer without finding a non-blank stmt"))
  (let ((initial-pos (point))
	(initial-indent (current-indentation))
	last-pos			; position of last stmt in region
	(followers
	 '((if elif else) (elif elif else) (else)
	   (try except finally) (except except) (finally)
	   (for else) (while else)
	   (def) (class) ) )
	first-symbol next-symbol)

    (cond
     ;; if comment line, suck up the following comment lines
     ((looking-at "[ \t]*#")
      (re-search-forward "^[ \t]*[^ \t#]" nil 'move) ; look for non-comment
      (re-search-backward "^[ \t]*#")	; and back to last comment in block
      (setq last-pos (point)))

     ;; else if line is a block line and EXTEND given, suck up
     ;; the whole structure
     ((and extend
	   (setq first-symbol (qmath-suck-up-first-keyword) )
	   (assq first-symbol followers))
      (while (and
	      (or (qmath-goto-beyond-block) t) ; side effect
	      (forward-line -1)		; side effect
	      (setq last-pos (point))	; side effect
	      (qmath-goto-statement-below)
	      (= (current-indentation) initial-indent)
	      (setq next-symbol (qmath-suck-up-first-keyword))
	      (memq next-symbol (cdr (assq first-symbol followers))))
	(setq first-symbol next-symbol)))

     ;; else if line *opens* a block, search for next stmt indented <=
     ((qmath-statement-opens-block-p)
      (while (and
	      (setq last-pos (point))	; always true -- side effect
	      (qmath-goto-statement-below)
	      (> (current-indentation) initial-indent))
	nil))

     ;; else plain code line; stop at next blank line, or stmt or
     ;; indenting comment line indented <
     (t
      (while (and
	      (setq last-pos (point))	; always true -- side effect
	      (or (qmath-goto-beyond-final-line) t)
	      (not (looking-at "[ \t]*$")) ; stop at blank line
	      (or
	       (>= (current-indentation) initial-indent)
	       (looking-at "[ \t]*#[^ \t\n]"))) ; ignore non-indenting #
	nil)))

    ;; skip to end of last stmt
    (goto-char last-pos)
    (qmath-goto-beyond-final-line)

    ;; set mark & display
    (if just-move
	()				; just return
      (push-mark (point) 'no-msg)
      (forward-line -1)
      (message "Mark set after: %s" (qmath-suck-up-leading-text))
      (goto-char initial-pos))))

(defun qmath-mark-def-or-class (&optional class)
  "Set region to body of def (or class, with prefix arg) enclosing point.
Pushes the current mark, then point, on the mark ring (all language
modes do this, but although it's handy it's never documented ...).

In most Emacs language modes, this function bears at least a
hallucinogenic resemblance to `\\[qmath-end-of-def-or-class]' and
`\\[qmath-beginning-of-def-or-class]'.

And in earlier versions of QMath mode, all 3 were tightly connected.
Turned out that was more confusing than useful: the `goto start' and
`goto end' commands are usually used to search through a file, and
people expect them to act a lot like `search backward' and `search
forward' string-search commands.  But because QMath `def' and `class'
can nest to arbitrary levels, finding the smallest def containing
point cannot be done via a simple backward search: the def containing
point may not be the closest preceding def, or even the closest
preceding def that's indented less.  The fancy algorithm required is
appropriate for the usual uses of this `mark' command, but not for the
`goto' variations.

So the def marked by this command may not be the one either of the
`goto' commands find: If point is on a blank or non-indenting comment
line, moves back to start of the closest preceding code statement or
indenting comment line.  If this is a `def' statement, that's the def
we use.  Else searches for the smallest enclosing `def' block and uses
that.  Else signals an error.

When an enclosing def is found: The mark is left immediately beyond
the last line of the def block.  Point is left at the start of the
def, except that: if the def is preceded by a number of comment lines
followed by (at most) one optional blank line, point is left at the
start of the comments; else if the def is preceded by a blank line,
point is left at its start.

The intent is to mark the containing def/class and its associated
documentation, to make moving and duplicating functions and classes
pleasant."
  (interactive "P")			; raw prefix arg
  (let ((start (point))
	(which (cond ((eq class 'either) "\\(class\\|def\\)")
		     (class "class")
		     (t "def"))))
    (push-mark start)
    (if (not (qmath-go-up-tree-to-keyword which))
	(progn (goto-char start)
	       (error "Enclosing %s not found"
		      (if (eq class 'either)
			  "def or class"
			which)))
      ;; else enclosing def/class found
      (setq start (point))
      (qmath-goto-beyond-block)
      (push-mark (point))
      (goto-char start)
      (if (zerop (forward-line -1))	; if there is a preceding line
	  (progn
	    (if (looking-at "[ \t]*$")	; it's blank
		(setq start (point))	; so reset start point
	      (goto-char start))	; else try again
	    (if (zerop (forward-line -1))
		(if (looking-at "[ \t]*#") ; a comment
		    ;; look back for non-comment line
		    ;; tricky: note that the regexp matches a blank
		    ;; line, cuz \n is in the 2nd character class
		    (and
		     (re-search-backward "^[ \t]*[^ \t#]" nil 'move)
		     (forward-line 1))
		  ;; no comment, so go back
		  (goto-char start)))))))
  (exchange-point-and-mark)
  (qmath-keep-region-active))

;; ripped from cc-mode
(defun qmath-forward-into-nomenclature (&optional arg)
  "Move forward to end of a nomenclature section or word.
With \\[universal-argument] (programmatically, optional argument ARG), 
do it that many times.

A `nomenclature' is a fancy way of saying AWordWithMixedCaseNotUnderscores."
  (interactive "p")
  (let ((case-fold-search nil))
    (if (> arg 0)
	(re-search-forward
	 "\\(\\W\\|[_]\\)*\\([A-Z]*[a-z0-9]*\\)"
	 (point-max) t arg)
      (while (and (< arg 0)
		  (re-search-backward
		   "\\(\\W\\|[a-z0-9]\\)[A-Z]+\\|\\(\\W\\|[_]\\)\\w+"
		   (point-min) 0))
	(forward-char 1)
	(setq arg (1+ arg)))))
  (qmath-keep-region-active))

(defun qmath-backward-into-nomenclature (&optional arg)
  "Move backward to beginning of a nomenclature section or word.
With optional ARG, move that many times.  If ARG is negative, move
forward.

A `nomenclature' is a fancy way of saying AWordWithMixedCaseNotUnderscores."
  (interactive "p")
  (qmath-forward-into-nomenclature (- arg))
  (qmath-keep-region-active))



;; Documentation functions

;; dump the long form of the mode blurb; does the usual doc escapes,
;; plus lines of the form ^[vc]:name$ to suck variable & command docs
;; out of the right places, along with the keys they're on & current
;; values
(defun qmath-dump-help-string (str)
  (with-output-to-temp-buffer "*Help*"
    (let ((locals (buffer-local-variables))
	  funckind funcname func funcdoc
	  (start 0) mstart end
	  keys )
      (while (string-match "^%\\([vc]\\):\\(.+\\)\n" str start)
	(setq mstart (match-beginning 0)  end (match-end 0)
	      funckind (substring str (match-beginning 1) (match-end 1))
	      funcname (substring str (match-beginning 2) (match-end 2))
	      func (intern funcname))
	(princ (substitute-command-keys (substring str start mstart)))
	(cond
	 ((equal funckind "c")		; command
	  (setq funcdoc (documentation func)
		keys (concat
		      "Key(s): "
		      (mapconcat 'key-description
				 (where-is-internal func qmath-mode-map)
				 ", "))))
	 ((equal funckind "v")		; variable
	  (setq funcdoc (documentation-property func 'variable-documentation)
		keys (if (assq func locals)
			 (concat
			  "Local/Global values: "
			  (prin1-to-string (symbol-value func))
			  " / "
			  (prin1-to-string (default-value func)))
		       (concat
			"Value: "
			(prin1-to-string (symbol-value func))))))
	 (t				; unexpected
	  (error "Error in qmath-dump-help-string, tag `%s'" funckind)))
	(princ (format "\n-> %s:\t%s\t%s\n\n"
		       (if (equal funckind "c") "Command" "Variable")
		       funcname keys))
	(princ funcdoc)
	(terpri)
	(setq start end))
      (princ (substitute-command-keys (substring str start))))
    (print-help-return-message)))

(defun qmath-describe-mode ()
  "Dump long form of QMath-mode docs."
  (interactive)
  (qmath-dump-help-string "Major mode for editing QMath files.
This documentation is still the pyhon-mode one, so some parts make sense,
and other ones do not. Read at your own risk !!

Knows about QMath indentation, tokens, comments and continuation lines.
Paragraphs are separated by blank lines only.

Major sections below begin with the string `@'; specific function and
variable docs begin with `->'.

@EXECUTING QMATH CODE

\\[qmath-execute-import-or-reload]\timports or reloads the file in the QMath interpreter
\\[qmath-execute-buffer]\tsends the entire buffer to the QMath interpreter
\\[qmath-execute-region]\tsends the current region
\\[qmath-execute-def-or-class]\tsends the current function or class definition
\\[qmath-execute-string]\tsends an arbitrary string
\\[qmath-shell]\tstarts a QMath interpreter window; this will be used by
\tsubsequent QMath execution commands
%c:qmath-execute-import-or-reload
%c:qmath-execute-buffer
%c:qmath-execute-region
%c:qmath-execute-def-or-class
%c:qmath-execute-string
%c:qmath-shell

@VARIABLES

qmath-indent-offset\tindentation increment
qmath-block-comment-prefix\tcomment string used by comment-region

qmath-qmath-command\tshell command to invoke QMath interpreter
qmath-temp-directory\tdirectory used for temp files (if needed)

qmath-beep-if-tab-change\tring the bell if tab-width is changed
%v:qmath-indent-offset
%v:qmath-block-comment-prefix
%v:qmath-qmath-command
%v:qmath-temp-directory
%v:qmath-beep-if-tab-change

@KINDS OF LINES

Each physical line in the file is either a `continuation line' (the
preceding line ends with a backslash that's not part of a comment, or
the paren/bracket/brace nesting level at the start of the line is
non-zero, or both) or an `initial line' (everything else).

An initial line is in turn a `blank line' (contains nothing except
possibly blanks or tabs), a `comment line' (leftmost non-blank
character is `#'), or a `code line' (everything else).

Comment Lines

Although all comment lines are treated alike by QMath, QMath mode
recognizes two kinds that act differently with respect to indentation.

An `indenting comment line' is a comment line with a blank, tab or
nothing after the initial `#'.  The indentation commands (see below)
treat these exactly as if they were code lines: a line following an
indenting comment line will be indented like the comment line.  All
other comment lines (those with a non-whitespace character immediately
following the initial `#') are `non-indenting comment lines', and
their indentation is ignored by the indentation commands.

Indenting comment lines are by far the usual case, and should be used
whenever possible.  Non-indenting comment lines are useful in cases
like these:

\ta = b   # a very wordy single-line comment that ends up being
\t        #... continued onto another line

\tif a == b:
##\t\tprint 'panic!' # old code we've `commented out'
\t\treturn a

Since the `#...' and `##' comment lines have a non-whitespace
character following the initial `#', QMath mode ignores them when
computing the proper indentation for the next line.

Continuation Lines and Statements

The QMath-mode commands generally work on statements instead of on
individual lines, where a `statement' is a comment or blank line, or a
code line and all of its following continuation lines (if any)
considered as a single logical unit.  The commands in this mode
generally (when it makes sense) automatically move to the start of the
statement containing point, even if point happens to be in the middle
of some continuation line.


@INDENTATION

Primarily for entering new code:
\t\\[indent-for-tab-command]\t indent line appropriately
\t\\[qmath-newline-and-indent]\t insert newline, then indent
\t\\[qmath-electric-backspace]\t reduce indentation, or delete single character

Primarily for reindenting existing code:
\t\\[qmath-guess-indent-offset]\t guess qmath-indent-offset from file content; change locally
\t\\[universal-argument] \\[qmath-guess-indent-offset]\t ditto, but change globally

\t\\[qmath-indent-region]\t reindent region to match its context
\t\\[qmath-shift-region-left]\t shift region left by qmath-indent-offset
\t\\[qmath-shift-region-right]\t shift region right by qmath-indent-offset

Unlike most programming languages, QMath uses indentation, and only
indentation, to specify block structure.  Hence the indentation supplied
automatically by QMath-mode is just an educated guess:  only you know
the block structure you intend, so only you can supply correct
indentation.

The \\[indent-for-tab-command] and \\[qmath-newline-and-indent] keys try to suggest plausible indentation, based on
the indentation of preceding statements.  E.g., assuming
qmath-indent-offset is 4, after you enter
\tif a > 0: \\[qmath-newline-and-indent]
the cursor will be moved to the position of the `_' (_ is not a
character in the file, it's just used here to indicate the location of
the cursor):
\tif a > 0:
\t    _
If you then enter `c = d' \\[qmath-newline-and-indent], the cursor will move
to
\tif a > 0:
\t    c = d
\t    _
QMath-mode cannot know whether that's what you intended, or whether
\tif a > 0:
\t    c = d
\t_
was your intent.  In general, QMath-mode either reproduces the
indentation of the (closest code or indenting-comment) preceding
statement, or adds an extra qmath-indent-offset blanks if the preceding
statement has `:' as its last significant (non-whitespace and non-
comment) character.  If the suggested indentation is too much, use
\\[qmath-electric-backspace] to reduce it.

Continuation lines are given extra indentation.  If you don't like the
suggested indentation, change it to something you do like, and QMath-
mode will strive to indent later lines of the statement in the same way.

If a line is a continuation line by virtue of being in an unclosed
paren/bracket/brace structure (`list', for short), the suggested
indentation depends on whether the current line contains the first item
in the list.  If it does, it's indented qmath-indent-offset columns beyond
the indentation of the line containing the open bracket.  If you don't
like that, change it by hand.  The remaining items in the list will mimic
whatever indentation you give to the first item.

If a line is a continuation line because the line preceding it ends with
a backslash, the third and following lines of the statement inherit their
indentation from the line preceding them.  The indentation of the second
line in the statement depends on the form of the first (base) line:  if
the base line is an assignment statement with anything more interesting
than the backslash following the leftmost assigning `=', the second line
is indented two columns beyond that `='.  Else it's indented to two
columns beyond the leftmost solid chunk of non-whitespace characters on
the base line.

Warning:  indent-region should not normally be used!  It calls \\[indent-for-tab-command]
repeatedly, and as explained above, \\[indent-for-tab-command] can't guess the block
structure you intend.
%c:indent-for-tab-command
%c:qmath-newline-and-indent
%c:qmath-electric-backspace


The next function may be handy when editing code you didn't write:
%c:qmath-guess-indent-offset


The remaining `indent' functions apply to a region of QMath code.  They
assume the block structure (equals indentation, in QMath) of the region
is correct, and alter the indentation in various ways while preserving
the block structure:
%c:qmath-indent-region
%c:qmath-shift-region-left
%c:qmath-shift-region-right

@MARKING & MANIPULATING REGIONS OF CODE

\\[qmath-mark-block]\t mark block of lines
\\[qmath-mark-def-or-class]\t mark smallest enclosing def
\\[universal-argument] \\[qmath-mark-def-or-class]\t mark smallest enclosing class
\\[comment-region]\t comment out region of code
\\[universal-argument] \\[comment-region]\t uncomment region of code
%c:qmath-mark-block
%c:qmath-mark-def-or-class
%c:comment-region

@MOVING POINT

\\[qmath-previous-statement]\t move to statement preceding point
\\[qmath-next-statement]\t move to statement following point
\\[qmath-goto-block-up]\t move up to start of current block
\\[qmath-beginning-of-def-or-class]\t move to start of def
\\[universal-argument] \\[qmath-beginning-of-def-or-class]\t move to start of class
\\[qmath-end-of-def-or-class]\t move to end of def
\\[universal-argument] \\[qmath-end-of-def-or-class]\t move to end of class

The first two move to one statement beyond the statement that contains
point.  A numeric prefix argument tells them to move that many
statements instead.  Blank lines, comment lines, and continuation lines
do not count as `statements' for these commands.  So, e.g., you can go
to the first code statement in a file by entering
\t\\[beginning-of-buffer]\t to move to the top of the file
\t\\[qmath-next-statement]\t to skip over initial comments and blank lines
Or do `\\[qmath-previous-statement]' with a huge prefix argument.
%c:qmath-previous-statement
%c:qmath-next-statement
%c:qmath-goto-block-up
%c:qmath-beginning-of-def-or-class
%c:qmath-end-of-def-or-class

@LITTLE-KNOWN EMACS COMMANDS PARTICULARLY USEFUL IN QMATH MODE

`\\[indent-new-comment-line]' is handy for entering a multi-line comment.

`\\[set-selective-display]' with a `small' prefix arg is ideally suited for viewing the
overall class and def structure of a module.

`\\[back-to-indentation]' moves point to a line's first non-blank character.

`\\[indent-relative]' is handy for creating odd indentation.

@OTHER EMACS HINTS

If you don't like the default value of a variable, change its value to
whatever you do like by putting a `setq' line in your .emacs file.
E.g., to set the indentation increment to 4, put this line in your
.emacs:
\t(setq  qmath-indent-offset  4)
To see the value of a variable, do `\\[describe-variable]' and enter the variable
name at the prompt.

When entering a key sequence like `C-c C-n', it is not necessary to
release the CONTROL key after doing the `C-c' part -- it suffices to
press the CONTROL key, press and release `c' (while still holding down
CONTROL), press and release `n' (while still holding down CONTROL), &
then release CONTROL.

Entering QMath mode calls with no arguments the value of the variable
`qmath-mode-hook', if that value exists and is not nil; for backward
compatibility it also tries `qmath-mode-hook'; see the `Hooks' section of
the Elisp manual for details.

Obscure:  When qmath-mode is first loaded, it looks for all bindings
to newline-and-indent in the global keymap, and shadows them with
local bindings to qmath-newline-and-indent."))


;; Helper functions
(defvar qmath-parse-state-re
  (concat
   "^[ \t]*\\(\\$\\|[a-zA-Z_0-9]:\\)\\>"
   "\\|"
   "^[^ #\t\n]"))

(defun qmath-parse-state ()
  "Return the parse state at point (see `parse-partial-sexp' docs)."
  (save-excursion
    (let ((here (point))
	  pps done)
      (while (not done)
	;; back up to the first preceding line (if any; else start of
	;; buffer) that begins with a popular QMath keyword, or a
	;; non- whitespace and non-comment character.  These are good
	;; places to start parsing to see whether where we started is
	;; at a non-zero nesting level.  It may be slow for people who
	;; write huge code blocks or huge lists ... tough beans.
	(re-search-backward qmath-parse-state-re nil 'move)
	(beginning-of-line)
	;; In XEmacs, we have a much better way to test for whether
	;; we're in a triple-quoted string or not.  Emacs does not
	;; have this built-in function, which is its loss because
	;; without scanning from the beginning of the buffer, there's
	;; no accurate way to determine this otherwise.
	(if (not (fboundp 'buffer-syntactic-context))
	    ;; Emacs
	    (progn
	      (save-excursion (setq pps (parse-partial-sexp (point) here)))
	      ;; make sure we don't land inside a triple-quoted string
	      (setq done (or (not (nth 3 pps))
			     (bobp)))
	      ;; Just go ahead and short circuit the test back to the
	      ;; beginning of the buffer.  This will be slow, but not
	      ;; nearly as slow as looping through many
	      ;; re-search-backwards.
	      (if (not done)
		  (goto-char (point-min))))
	  ;; XEmacs
	  (setq done (or (not (buffer-syntactic-context))
			 (bobp)))
	  (when done
	    (setq pps (parse-partial-sexp (point) here)))
	  ))
      pps)))

(defun qmath-nesting-level ()
  "Return the buffer position of the last unclosed enclosing list.
If nesting level is zero, return nil."
  (let ((status (qmath-parse-state)))
    (if (zerop (car status))
	nil				; not in a nest
      (car (cdr status)))))		; char# of open bracket

(defun qmath-backslash-continuation-line-p ()
  "Return t iff preceding line ends with backslash that is not in a comment."
  (save-excursion
    (beginning-of-line)
    (and
     ;; use a cheap test first to avoid the regexp if possible
     ;; use 'eq' because char-after may return nil
     (eq (char-after (- (point) 2)) ?\\ )
     ;; make sure; since eq test passed, there is a preceding line
     (forward-line -1)			; always true -- side effect
     (looking-at qmath-continued-re))))

(defun qmath-continuation-line-p ()
  "Return t iff current line is a continuation line."
  (save-excursion
    (beginning-of-line)
    (or (qmath-backslash-continuation-line-p)
	(qmath-nesting-level))))

(defun qmath-goto-beginning-of-tqs (delim)
  "Go to the beginning of the triple quoted string we find ourselves in.
DELIM is the TQS string delimiter character we're searching backwards
for."
  (let ((skip (and delim (make-string 1 delim))))
    (when skip
      (save-excursion
	(qmath-safe (search-backward skip))
	(if (and (eq (char-before) delim)
		 (eq (char-before (1- (point))) delim))
	    (setq skip (make-string 3 delim))))
      ;; we're looking at a triple-quoted string
      (qmath-safe (search-backward skip)))))

(defun qmath-goto-initial-line ()
  "Go to the initial line of the current statement.
Usually this is the line we're on, but if we're on the 2nd or
following lines of a continuation block, we need to go up to the first
line of the block."
  ;; Tricky: We want to avoid quadratic-time behavior for long
  ;; continued blocks, whether of the backslash or open-bracket
  ;; varieties, or a mix of the two.  The following manages to do that
  ;; in the usual cases.
  ;;
  ;; Also, if we're sitting inside a triple quoted string, this will
  ;; drop us at the line that begins the string.
  (let (open-bracket-pos)
    (while (qmath-continuation-line-p)
      (beginning-of-line)
      (if (qmath-backslash-continuation-line-p)
	  (while (qmath-backslash-continuation-line-p)
	    (forward-line -1))
	;; else zip out of nested brackets/braces/parens
	(while (setq open-bracket-pos (qmath-nesting-level))
	  (goto-char open-bracket-pos)))))
  (beginning-of-line))

(defun qmath-goto-beyond-final-line ()
  "Go to the point just beyond the fine line of the current statement.
Usually this is the start of the next line, but if this is a
multi-line statement we need to skip over the continuation lines."
  ;; Tricky: Again we need to be clever to avoid quadratic time
  ;; behavior.
  ;;
  ;; XXX: Not quite the right solution, but deals with multi-line doc
  ;; strings
  (if (looking-at (concat "[ \t]*\\(" qmath-stringlit-re "\\)"))
      (goto-char (match-end 0)))
  ;;
  (forward-line 1)
  (let (state)
    (while (and (qmath-continuation-line-p)
		(not (eobp)))
      ;; skip over the backslash flavor
      (while (and (qmath-backslash-continuation-line-p)
		  (not (eobp)))
	(forward-line 1))
      ;; if in nest, zip to the end of the nest
      (setq state (qmath-parse-state))
      (if (and (not (zerop (car state)))
	       (not (eobp)))
	  (progn
	    (parse-partial-sexp (point) (point-max) 0 nil state)
	    (forward-line 1))))))

(defun qmath-statement-opens-block-p ()
  "Return t iff the current statement opens a block.
I.e., iff it ends with a colon that is not in a comment.  Point should 
be at the start of a statement."
  (save-excursion
    (let ((start (point))
	  (finish (progn (qmath-goto-beyond-final-line) (1- (point))))
	  (searching t)
	  (answer nil)
	  state)
      (goto-char start)
      (while searching
	;; look for a colon with nothing after it except whitespace, and
	;; maybe an identifier
	(if (re-search-forward ":\\([ \t]\\)*\\(\\[[a-zA-Z_<-]+\\]\\)?\\([ \t]\\)*$"
			       finish t)
	    (if (eq (point) finish)	; note: no `else' clause; just
					; keep searching if we're not at
					; the end yet
		;; sure looks like it opens a block -- but it might
		;; be in a comment
		(progn
		  (setq searching nil)	; search is done either way
		  (setq state (parse-partial-sexp start
						  (match-beginning 0)))
		  (setq answer (not (nth 4 state)))))
	  ;; search failed: couldn't find another interesting colon
	  (setq searching nil)))
      answer)))

(defun qmath-statement-closes-block-p ()
  "Return t iff the current statement closes a block.
I.e., if the line starts with `return', `raise', `break', `continue',
and `pass'.  This doesn't catch embedded statements."
  (let ((here (point)))
    (qmath-goto-initial-line)
    (back-to-indentation)
    (prog1
	(looking-at (concat qmath-block-closing-keywords-re "\\>"))
      (goto-char here))))

(defun qmath-goto-beyond-block ()
  "Go to point just beyond the final line of block begun by the current line.
This is the same as where `qmath-goto-beyond-final-line' goes unless
we're on colon line, in which case we go to the end of the block.
Assumes point is at the beginning of the line."
  (if (qmath-statement-opens-block-p)
      (qmath-mark-block nil 'just-move)
    (qmath-goto-beyond-final-line)))

(defun qmath-goto-statement-at-or-above ()
  "Go to the start of the first statement at or preceding point.
Return t if there is such a statement, otherwise nil.  `Statement'
does not include blank lines, comments, or continuation lines."
  (qmath-goto-initial-line)
  (if (looking-at qmath-blank-or-comment-re)
      ;; skip back over blank & comment lines
      ;; note:  will skip a blank or comment line that happens to be
      ;; a continuation line too
      (if (re-search-backward "^[ \t]*[^ \t#\n]" nil t)
	  (progn (qmath-goto-initial-line) t)
	nil)
    t))

(defun qmath-goto-statement-below ()
  "Go to start of the first statement following the statement containing point.
Return t if there is such a statement, otherwise nil.  `Statement'
does not include blank lines, comments, or continuation lines."
  (beginning-of-line)
  (let ((start (point)))
    (qmath-goto-beyond-final-line)
    (while (and
	    (looking-at qmath-blank-or-comment-re)
	    (not (eobp)))
      (forward-line 1))
    (if (eobp)
	(progn (goto-char start) nil)
      t)))

(defun qmath-go-up-tree-to-keyword (key)
  "Go to begining of statement starting with KEY, at or preceding point.

KEY is a regular expression describing a QMath keyword.  Skip blank
lines and non-indenting comments.  If the statement found starts with
KEY, then stop, otherwise go back to first enclosing block starting
with KEY.  If successful, leave point at the start of the KEY line and 
return t.  Otherwise, leav point at an undefined place and return nil."
  ;; skip blanks and non-indenting #
  (qmath-goto-initial-line)
  (while (and
	  (looking-at "[ \t]*\\($\\|#[^ \t\n]\\)")
	  (zerop (forward-line -1)))	; go back
    nil)
  (qmath-goto-initial-line)
  (let* ((re (concat "[ \t]*" key "\\b"))
	 (case-fold-search nil)		; let* so looking-at sees this
	 (found (looking-at re))
	 (dead nil))
    (while (not (or found dead))
      (condition-case nil		; in case no enclosing block
	  (qmath-goto-block-up 'no-mark)
	(error (setq dead t)))
      (or dead (setq found (looking-at re))))
    (beginning-of-line)
    found))

(defun qmath-suck-up-leading-text ()
  "Return string in buffer from start of indentation to end of line.
Prefix with \"...\" if leading whitespace was skipped."
  (save-excursion
    (back-to-indentation)
    (concat
     (if (bolp) "" "...")
     (buffer-substring (point) (progn (end-of-line) (point))))))

(defun qmath-suck-up-first-keyword ()
  "Return first keyword on the line as a Lisp symbol.
`Keyword' is defined (essentially) as the regular expression
([a-z]+).  Returns nil if none was found."
  (let ((case-fold-search nil))
    (if (looking-at "[ \t]*\\([a-z]+\\)\\b")
	(intern (buffer-substring (match-beginning 1) (match-end 1)))
      nil)))

(defun qmath-current-defun ()
  "QMath value for `add-log-current-defun-function'.
This tells add-log.el how to find the current function/method/variable."
  (save-excursion
    (if (re-search-backward qmath-defun-start-re nil t)
	(or (match-string 3)
	    (let ((method (match-string 2)))
	      (if (and (not (zerop (length (match-string 1))))
		       (re-search-backward qmath-class-start-re nil t))
		  (concat (match-string 1) "." method)
		method)))
      nil)))


(defconst qmath-help-address "matmota@matracas.org"
  "Address accepting submission of bug reports.")

(defun qmath-version ()
  "Echo the current version of `qmath-mode' in the minibuffer."
  (interactive)
  (message "Using `qmath-mode' version %s" qmath-version)
  (qmath-keep-region-active))

;; only works under Emacs 19
;(eval-when-compile
;  (require 'reporter))

(defun qmath-submit-bug-report (enhancement-p)
  "Submit via mail a bug report on `qmath-mode'.
With \\[universal-argument] (programmatically, argument ENHANCEMENT-P
non-nil) just submit an enhancement request."
  (interactive
   (list (not (y-or-n-p
	       "Is this a bug report (hit `n' to send other comments)? "))))
  (let ((reporter-prompt-for-summary-p (if enhancement-p
					   "(Very) brief summary: "
					 t)))
    (require 'reporter)
    (reporter-submit-bug-report
     qmath-help-address			;address
     (concat "qmath-mode " qmath-version)	;pkgname
     ;; varlist
     (if enhancement-p nil
       '(qmath-qmath-command
	 qmath-indent-offset
	 qmath-block-comment-prefix
	 qmath-temp-directory
	 qmath-beep-if-tab-change))
     nil				;pre-hooks
     nil				;post-hooks
     "Dear Bug Ironer,")		;salutation
    (if enhancement-p nil
      (set-mark (point))
      (insert 
"Please replace this text with a sufficiently large code sample\n\
and an exact recipe so that I can reproduce your problem.  Failure\n\
to do so may mean a greater delay in fixing your bug.\n\n")
      (exchange-point-and-mark)
      (qmath-keep-region-active))))


(defun qmath-kill-emacs-hook ()
  "Delete files in `qmath-file-queue'.
These are QMath temporary files awaiting execution."
  (mapcar #'(lambda (filename)
	      (qmath-safe (delete-file filename)))
	  qmath-file-queue))

;; arrange to kill temp files when Emacs exists
(add-hook 'kill-emacs-hook 'qmath-kill-emacs-hook)



(provide 'qmath-mode)
;;; qmath-mode.el ends here
