// Clase:    OpenMath_object
// Proyecto: QMath
// Fecha:    jueves, 17 de diciembre de 2009, 22:56:38 CET
///////////////////////////////////////////////////////////////////////////
//
//   Copyright  2001, 2002 Alberto Gonzlez Palomo
//   Author: Alberto Gonzlez Palomo
//
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
//
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program; if not, write to the Free Software
//   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
/////////////////////////////////////////////////////////////////////////////


#include "OpenMath_object.h"

/////////////
// OpenMath_object - Class constructor.
// no arguments.
//public:
OpenMath_object::OpenMath_object()
{
  this->basic_object = NULL;
  this->xml_string = NULL;
  this->construct = NULL;
  Type(none);
}

/////////////
// ~OpenMath_object - Class destructor.
// no arguments.
//public:
OpenMath_object::~OpenMath_object()
{
  Type(none);// This takes care of object deletion.
}

using std::list;

/////////////
// Binding - Append objects to an OpenMath binding object.
// num_args: Number of objects to append.
// returns object pointer for operator concatenation.
//public:
OpenMath_object_p_t OpenMath_object::Binding(int num_args, ...)
{
  Type(OpenMath_object::binding);
  if (num_args < 1) return this;
  
  va_list arg_list;
  va_start(arg_list, num_args);
  for (int arg_index = 1; arg_index <= num_args; ++arg_index)
    {
      construct->push_back(va_arg(arg_list, OpenMath_object_p_t));
    }
  va_end(arg_list);
  
  return this;
}


/////////////
// Binding - Append objects to an OpenMath binding object.
// object_list: const list<OpenMath_object_p_t>* of objects to append.
// returns object pointer for operator concatenation.
//public:
OpenMath_object_p_t OpenMath_object::Binding(const list<OpenMath_object_p_t>* object_list)
{
  Type(OpenMath_object::binding);
  if (!object_list) return this;
  if (object_list->size() < 1) return this;
  
  list<OpenMath_object_p_t>::const_iterator object_list_iterator;
  for (object_list_iterator = object_list->begin();
       object_list_iterator != object_list->end();
       ++object_list_iterator)
    {
      construct->push_back(*object_list_iterator);
    }
  
  return this;
}

/////////////
// operator<< - Overloading of << operator.
// output: output c++ stream.
// object: OpenMath object to output.
// returns output stream for operator composition.
//public:
ostream &operator<< (ostream &output, OpenMath_object_p_t object)
{
  object->Output(object, output);
  
  return output;
}


/////////////
// Byte_array - Set OpenMath byte_array object value.
// byte_array: new value, encoded as a zero-terminated C ASCII string.
// returns object pointer for operator concatenation.
//public:
OpenMath_object_p_t OpenMath_object::Byte_array(char* byte_array, int count)
{
  if (!byte_array) return this;
  Type(OpenMath_object::byte_array);
  
  basic_object->erase(basic_object->begin(), basic_object->end());
  for (int i = 0; i < count; ++i)
    basic_object->push_back((unsigned char)byte_array[i]);
  
  return this;
}

/////////////
// Byte_array - Access to OpenMath byte_array object.
// no arguments.
// returns pointer to char vector.
//public:
vector<unsigned char>* OpenMath_object::Byte_array()
{
  return this->basic_object;
}


/////////////
// Output - Write to a stream in XML format.
// item: OpenMath_object to be output, passed as pointer.
// output: stream which receives data.
// returns true on success, false on fail.
//public:
//static
bool OpenMath_object::Output(OpenMath_object_p_t object, ostream &output)
{
  if (!object || !(&output)) return false;
  
  bool return_value;
  
  output << "<OMOBJ>";
  return_value = Output_subobject(object, output);
  output << "</OMOBJ>";
  
  return return_value;
}

/////////////
// Output_construct_objects - Write a list of objects.
// construct_iterator: the list<OpenMath_object_p_t> current iterator, passed
//  by reference. Gets modified in place.
// count: iteration count.
// construct_end: end iterator of list<OpenMath_object_p_t>.
//private:
//static
void OpenMath_object::Output_construct_objects(ostream &output, list<OpenMath_object_p_t>::iterator &construct_iterator, int count, list<OpenMath_object_p_t>::iterator &construct_end)
{
  for (int index = 0; index < count; ++index)
    {
      if (construct_iterator == construct_end) break;
      if (*construct_iterator != NULL)
        Output_subobject(*(construct_iterator++), output);
      else
        output << "«NULL»";
    }
}

/////////////
// Output_subobject - Write to a stream in XML format.
// item: OpenMath_object to be output, passed as pointer.
// output: stream which receives data.
// returns true on success, false on fail.
//private:
//static
bool OpenMath_object::Output_subobject(OpenMath_object_p_t object, ostream &output)
{
  if (!object || !(&output)) return false;
  
  switch (object->Type())
    {
    case none:        output << "<NoNe>";            break;
    case integer:     output << "<OMI>";             break;
    case fp_number:   output << "<OMF";              break;
    case char_string: output << "<OMSTR>";           break;
    case byte_array:  output << "<OMB>";             break;
    case symbol:      output << "<OMS";              break;
    case variable:    output << "<OMV";              break;
    case application: output << "<OMA>";             break;
    case attribution: output << "<OMATTR>";          break;
    case binding:     output << "<OMBIND>";          break;
    case error:       output << "<OME>";             break;
    }
  
  vector<unsigned char>::iterator char_iterator;
  vector<unsigned char>::iterator separator_char;
  list<OpenMath_object_p_t>::iterator construct_iterator;
  list<OpenMath_object_p_t>::iterator construct_end;
  int construct_size;
  
  switch (object->Type())
    {
    case none:
      break;
    case integer:
      for (char_iterator = object->basic_object->begin();
           char_iterator < object->basic_object->end();
           ++char_iterator)
        output << *char_iterator;
      break;
    case symbol:
      separator_char = find(object->basic_object->begin(),
                            object->basic_object->end(),
                            ':');
      if (separator_char != object->basic_object->end())
        {
          output << " cd=\"";
          for (char_iterator = object->basic_object->begin();
               char_iterator < separator_char;
               ++char_iterator)
            output << *char_iterator;
          output << "\"";
          ++separator_char;
        }
      else
        {
          separator_char = object->basic_object->begin();
        }
      output << " name=\"";
      for (char_iterator = separator_char;
           char_iterator < object->basic_object->end();
           ++char_iterator)
        output << *char_iterator;
      output << "\"";
      break;
    case variable:
      output << " name=\"";
      for (char_iterator = object->basic_object->begin();
           char_iterator < object->basic_object->end();
           ++char_iterator)
        output << *char_iterator;
      output << "\"";
      break;
    case fp_number:
      output << " dec=\"";
      int previous_precision;
      previous_precision = output.precision();
      output.precision(16);
      output << object->double_float_number;
      output.precision(previous_precision);
      output << "\"";
      break;
    case char_string:
      output << object->xml_string;
      break;
    case byte_array:
      for (char_iterator = object->basic_object->begin();
           char_iterator < object->basic_object->end();
           ++char_iterator)
        // TODO: uuencode
        output << '@';//*char_iterator;
      break;
    case application:
      construct_iterator = object->construct->begin();
      construct_size = object->construct->size();
      construct_end = object->construct->end();
      Output_construct_objects(output, construct_iterator,
                               construct_size, construct_end);
      break;
    case attribution:
      construct_iterator = object->construct->begin();
      construct_size = object->construct->size();
      construct_end = object->construct->end();
      output << "<OMATP>";
      Output_construct_objects(output, construct_iterator,
                               construct_size - 1, construct_end);
      output << "</OMATP>";
      Output_construct_objects(output, construct_iterator,
                               1, construct_end);
      break;
    case binding:
      construct_iterator = object->construct->begin();
      construct_size = object->construct->size();
      construct_end = object->construct->end();
      Output_construct_objects(output, construct_iterator,
                               1, construct_end);
      output << "<OMBVAR>";
      Output_construct_objects(output, construct_iterator,
                               construct_size - 2, construct_end);
      output << "</OMBVAR>";
      Output_construct_objects(output, construct_iterator,
                               1, construct_end);
      break;
    case error:
      construct_iterator = object->construct->begin();
      construct_size = object->construct->size();
      construct_end = object->construct->end();
      Output_construct_objects(output, construct_iterator,
                               construct_size, construct_end);
      break;
    }
  
  switch (object->Type())
    {
    case none:        output << "</NoNe>";     break;
    case integer:     output << "</OMI>";      break;
    case fp_number:   output << "/>";          break;
    case char_string: output << "</OMSTR>";    break;
    case byte_array:  output << "</OMB>";      break;
    case symbol:      output << "/>";          break;
    case variable:    output << "/>";          break;
    case application: output << "</OMA>";      break;
    case attribution: output << "</OMATTR>";   break;
    case binding:     output << "</OMBIND>";   break;
    case error:       output << "</OME>";      break;
    }
  
  return true;
}

using std::vector;

/////////////
// Attribution - Append objects to an OpenMath attribution object.
// num_args: Number of objects to append.
// returns object pointer for operator concatenation.
//public:
OpenMath_object_p_t OpenMath_object::Attribution(int num_args, ...)
{
  Type(OpenMath_object::attribution);
  if (num_args < 1) return this;
  
  va_list arg_list;
  va_start(arg_list, num_args);
  for (int arg_index = 1; arg_index <= num_args; ++arg_index)
    {
      construct->push_back(va_arg(arg_list, OpenMath_object_p_t));
    }
  va_end(arg_list);
  
  return this;
}


/////////////
// Attribution - Append objects to an OpenMath attribution object.
// object_list: const list<OpenMath_object_p_t>* of objects to append.
// returns object pointer for operator concatenation.
//public:
OpenMath_object_p_t OpenMath_object::Attribution(const list<OpenMath_object_p_t>* object_list)
{
  Type(OpenMath_object::attribution);
  if (!object_list) return this;
  if (object_list->size() < 1) return this;
  
  list<OpenMath_object_p_t>::const_iterator object_list_iterator;
  for (object_list_iterator = object_list->begin();
       object_list_iterator != object_list->end();
       ++object_list_iterator)
    {
      construct->push_back(*object_list_iterator);
    }
  
  return this;
}

/////////////
// Type - Set OpenMath object type.
// type: new type for object.
//       If the change is from a construct type to the same construct type,
//       the construct list is preserved. This is needed for Application(),
//       Attribution(), Binding(), and Error() to append their arguments to
//       the existent construct instead of replacing them.
// returns old type.
//public:
OpenMath_object::OpenMath_object_type_t OpenMath_object::Type(OpenMath_object::OpenMath_object_type_t type)
{
  OpenMath_object_type_t old_type;
  old_type = this->type;
  //if (type == old_type) return old_type; Removed to ensure construct erasing
  
  switch (type)
    {
    case integer:
    case byte_array:
    case symbol:
    case variable:
      if (!basic_object) { basic_object = new vector<unsigned char>;       }
      if (xml_string)    { delete xml_string;     xml_string = NULL;       }
      if (construct)     { delete construct;       construct = NULL;       }
      this->type = type;
      break;
    case fp_number:
      if (basic_object)  { delete basic_object; basic_object = NULL;       }
      if (xml_string)    { delete xml_string;     xml_string = NULL;       }
      if (construct)     { delete construct;       construct = NULL;       }
      this->type = type;
      break;
    case char_string:
      if (basic_object)  { delete basic_object; basic_object = NULL;       }
      if (!xml_string)   { xml_string = new XML_string;                    }
      if (construct)     { delete construct;       construct = NULL;       }
      this->type = type;
      break;
    case application:
    case attribution:
    case binding:
    case error:
      if (basic_object)  { delete basic_object; basic_object = NULL;       }
      if (xml_string)    { delete xml_string;     xml_string = NULL;       }
      if (!construct)    { construct = new list<OpenMath_object_p_t>; }
      else
        {
          if (type != old_type)
            construct->erase(construct->begin(), construct->end());
        }
      this->type = type;
      break;
    default:
      // This is an error. Invalidate object. Someone will pay for this.
      if (basic_object)  { delete basic_object; basic_object = NULL;       }
      if (xml_string)    { delete xml_string;     xml_string = NULL;       }
      if (construct)     { delete construct;       construct = NULL;       }
      this->type = none;
      break;
    }
  return old_type;
}

/////////////
// Type - Report OpenMath object type.
// no arguments.
// returns current type.
//public:
OpenMath_object::OpenMath_object_type_t OpenMath_object::Type()
{
  return this->type;
}

using std::vector;


/////////////
// Char_string - Set char_string value from UCS4 char vector.
// xml_string: new value for property, encoded as UCS4.
// returns object pointer for operator concatenation.
//public:
OpenMath_object_p_t OpenMath_object::Char_string(vector<XML_UCS4_char_t> &xml_string)
{
  if (!(&xml_string)) return this;
  Type(OpenMath_object::char_string);
  if (!this->xml_string)
    this->xml_string = new XML_string();
  this->xml_string->UCS4(xml_string);
  return this;
}

/////////////
// Char_string - Set char_string value from UTF8 char vector.
// xml_string: new value for property, encoded as UTF8.
// returns object pointer for operator concatenation.
//public:
OpenMath_object_p_t OpenMath_object::Char_string(vector<char> &xml_string)
{
  if (!(&xml_string)) return this;
  Type(OpenMath_object::char_string);
  if (!this->xml_string)
    this->xml_string = new XML_string();
  this->xml_string->UTF8(xml_string);
  return this;
}

/////////////
// Char_string - Set char_string value from UTF8 char C++ string.
// xml_string: new value for property, encoded as UTF8.
// returns object pointer for operator concatenation.
//public:
OpenMath_object_p_t OpenMath_object::Char_string(string &xml_string)
{
  if (!(&xml_string)) return this;
  Type(OpenMath_object::char_string);
  if (!this->xml_string)
    this->xml_string = new XML_string();
  this->xml_string->UTF8(xml_string);
  return this;
}

/////////////
// Char_string - Set char string value from UTF8 char C array.
// xml_string: new value for property, encoded as UTF8.
// returns object pointer for operator concatenation.
//public:
OpenMath_object_p_t OpenMath_object::Char_string(const char* xml_string)
{
  if (!xml_string) return this;
  Type(OpenMath_object::char_string);
  if (!this->xml_string)
    this->xml_string = new XML_string();
  vector<char> xml_string_vector;
  for (unsigned int index = 0; index < strlen(xml_string); ++index)
    xml_string_vector.push_back(xml_string[index]);
  return Char_string(xml_string_vector);
}

/////////////
// Char_string - Access to char_string as an XML_string pointer.
// no arguments.
// returns pointer to xml_string.
//public:
XML_string_p_t OpenMath_object::Char_string()
{
  return this->xml_string;
}

using std::list;

/////////////
// Application - Append objects to an OpenMath application object.
// num_args: Number of objects to append.
// returns object pointer for operator concatenation.
//public:
OpenMath_object_p_t OpenMath_object::Application(int num_args, ...)
{
  Type(OpenMath_object::application);
  if (num_args < 1) return this;
  
  va_list arg_list;
  va_start(arg_list, num_args);
  for (int arg_index = 1; arg_index <= num_args; ++arg_index)
    {
      construct->push_back(va_arg(arg_list, OpenMath_object_p_t));
    }
  va_end(arg_list);
  
  return this;
}


/////////////
// Application - Append objects to an OpenMath application object.
// object_list: const list<OpenMath_object_p_t>* of objects to append.
// returns object pointer for operator concatenation.
//public:
OpenMath_object_p_t OpenMath_object::Application(const list<OpenMath_object_p_t>* object_list)
{
  Type(OpenMath_object::application);
  if (!object_list) return this;
  if (object_list->size() < 1) return this;
  
  list<OpenMath_object_p_t>::const_iterator object_list_iterator;
  for (object_list_iterator = object_list->begin();
       object_list_iterator != object_list->end();
       ++object_list_iterator)
    {
      construct->push_back(*object_list_iterator);
    }
  
  return this;
}

/////////////
// Symbol - Set OpenMath symbol object value.
// symbol: new value for object, encoded as an ASCII char vector.
// returns object pointer for operator concatenation.
//public:
OpenMath_object_p_t OpenMath_object::Symbol(vector<char> &symbol)
{
  if (!(&symbol)) return this;
  
  Type(OpenMath_object::symbol);
  basic_object->erase(basic_object->begin(), basic_object->end());
  basic_object->insert(basic_object->begin(),
                       symbol.begin(), symbol.end());
  
  return this;
}

/////////////
// Symbol - Set OpenMath symbol object value.
// symbol: new value for object, encoded as a zero-terminated C++ ASCII string.
// returns object pointer for operator concatenation.
//public:
OpenMath_object_p_t OpenMath_object::Symbol(string &symbol)
{
  if (!(&symbol)) return this;
  
  Type(OpenMath_object::symbol);
  basic_object->erase(basic_object->begin(), basic_object->end());
  basic_object->insert(basic_object->begin(),
                       symbol.begin(), symbol.end());
  
  return this;
}

/////////////
// Symbol - Set OpenMath symbol object value.
// symbol: new value for object, encoded as a zero-terminated C ASCII string.
// returns object pointer for operator concatenation.
//public:
OpenMath_object_p_t OpenMath_object::Symbol(const char* symbol)
{
  if (!symbol) return this;
  
  Type(OpenMath_object::symbol);
  
  return Symbol(symbol, strlen(symbol));
}

/////////////
// Symbol - Set OpenMath symbol object value.
// symbol: new value for object, encoded as a zero-terminated C ASCII string.
// returns object pointer for operator concatenation.
//public:
OpenMath_object_p_t OpenMath_object::Symbol(const char* symbol, int count)
{
  if (!symbol) return this;
  
  Type(OpenMath_object::symbol);
  basic_object->erase(basic_object->begin(), basic_object->end());
  for (int i = 0; i < count; ++i)
    basic_object->push_back((unsigned char)symbol[i]);
  
  return this;
}

/////////////
// Symbol - Access to OpenMath symbol object.
// no arguments.
// returns pointer to char vector.
//public:
vector<unsigned char>* OpenMath_object::Symbol()
{
  return this->basic_object;
}


/////////////
// Variable - Set OpenMath variable object value.
// variable: new value for object, encoded as an ASCII char vector.
// returns object pointer for operator concatenation.
//public:
OpenMath_object_p_t OpenMath_object::Variable(vector<char> &variable)
{
  if (!(&variable)) return this;
  
  Type(OpenMath_object::variable);
  basic_object->erase(basic_object->begin(), basic_object->end());
  basic_object->insert(basic_object->begin(),
                       variable.begin(), variable.end());
  
  return this;
}

/////////////
// Variable - Set OpenMath variable object value.
// variable: new value for object, encoded as a zero-terminated C++ ASCII string.
// returns object pointer for operator concatenation.
//public:
OpenMath_object_p_t OpenMath_object::Variable(string &variable)
{
  if (!(&variable)) return this;
  
  Type(OpenMath_object::variable);
  basic_object->erase(basic_object->begin(), basic_object->end());
  basic_object->insert(basic_object->begin(),
                       variable.begin(), variable.end());
  
  return this;
}

/////////////
// Variable - Set OpenMath variable object value.
// variable: new value for object, encoded as a zero-terminated C ASCII string.
// returns object pointer for operator concatenation.
//public:
OpenMath_object_p_t OpenMath_object::Variable(const char* variable)
{
  if (!variable) return this;
  
  Type(OpenMath_object::variable);
  
  return Variable(variable, strlen(variable));
}

/////////////
// Variable - Set OpenMath variable object value.
// variable: new value for object, encoded as a zero-terminated C ASCII string.
// returns object pointer for operator concatenation.
//public:
OpenMath_object_p_t OpenMath_object::Variable(const char* variable, int count)
{
  if (!variable) return this;
  
  Type(OpenMath_object::variable);
  basic_object->erase(basic_object->begin(), basic_object->end());
  for (int i = 0; i < count; ++i)
    basic_object->push_back((unsigned char)variable[i]);
  
  return this;
}

/////////////
// Variable - Access to OpenMath variable object.
// no arguments.
// returns pointer to char vector.
//public:
vector<unsigned char>* OpenMath_object::Variable()
{
  return this->basic_object;
}



/////////////
// To_XML_elem_info_item - Dump object to an XML_elem_info_item.
// no arguments.
// returns pointer to generated XML_elem_info_item.
//public:
XML_elem_info_item_p_t OpenMath_object::To_XML_elem_info_item()
{
  XML_elem_info_item_p_t item = NULL;
  
  item = new XML_elem_info_item;
  if (!item) return NULL;
  
  item->Local_name("OMOBJ")->Children(this->To_XML_elem_info_item_subobject());
  
  return item;
}

/////////////
// To_XML_elem_info_item_subobject - Dump object to an XML_elem_info_item.
// no arguments.
// returns pointer to generated XML_elem_info_item.
//private:
XML_elem_info_item_p_t OpenMath_object::To_XML_elem_info_item_subobject()
{
  XML_elem_info_item_p_t item = NULL;
  
  item = new XML_elem_info_item;
  if (!item) return NULL;
  
  switch (this->Type())
    {
    case none:        item->Local_name("NoNe");    break;
    case integer:     item->Local_name("OMI");     break;
    case fp_number:   item->Local_name("OMF");     break;
    case char_string: item->Local_name("OMSTR");   break;
    case byte_array:  item->Local_name("OMB");     break;
    case symbol:      item->Local_name("OMS");     break;
    case variable:    item->Local_name("OMV");     break;
    case application: item->Local_name("OMA");     break;
    case attribution: item->Local_name("OMATTR");  break;
    case binding:     item->Local_name("OMBIND");  break;
    case error:       item->Local_name("OME");     break;
    }
  
  vector<unsigned char>::iterator char_iterator;
  vector<unsigned char>::iterator separator_char;
  char* buffer = NULL;
  vector<char> char_buffer;
  list<OpenMath_object_p_t>::iterator construct_iterator;
  list<OpenMath_object_p_t>::iterator construct_end;
  int construct_size;
  
  switch (this->Type())
    {
    case none:
      break;
    case integer:
      char_buffer.insert(char_buffer.begin(),
                         this->basic_object->begin(),
                         this->basic_object->end());
      item->Children(char_buffer);
      break;
    case fp_number:
      buffer = (char*)malloc(SAFE_BUFFER_SIZE_FOR_DOUBLE_FLOAT_TO_STRING);
      if (!buffer) return item;
      sprintf(buffer, "%.16f", this->double_float_number);
      buffer[SAFE_BUFFER_SIZE_FOR_DOUBLE_FLOAT_TO_STRING - 1] = 0;
      for (unsigned int index = strlen(buffer); index > 2; --index)
        {
          if ('0' == buffer[index-1] && '0' == buffer[index-2])
            buffer[index-1] = 0;
          else
            break;
        }
      item->Attributes((new XML_attr_info_item)
                       ->Local_name("dec")
                       ->Normalized_value(buffer)
                       );
      break;
    case char_string:
      if (!xml_string) return item;
      item->Children(*xml_string);
      break;
    case byte_array:
      for (char_iterator = this->basic_object->begin();
           char_iterator < this->basic_object->end();
           ++char_iterator)
        {
          item->Children((new XML_character_info_item)
                         ->Parent(item)
                         // TODO: uuencode
                         ->Character_code(XML_UCS4_char_t('@'))
                         );
        }
      break;
    case symbol:
      separator_char = find(this->basic_object->begin(),
                            this->basic_object->end(),
                            ':');
      if (separator_char != this->basic_object->end())
        {
          char_buffer.insert(char_buffer.begin(),
                             this->basic_object->begin(), separator_char);
          item->Attributes((new XML_attr_info_item)
                           ->Local_name("cd")
                           ->Normalized_value(char_buffer)
                           );
          ++separator_char;
        }
      else
        {
          separator_char = this->basic_object->begin();
        }
      char_buffer.erase(char_buffer.begin(), char_buffer.end());
      char_buffer.insert(char_buffer.begin(),
                         separator_char, this->basic_object->end());
      item->Attributes((new XML_attr_info_item)
                       ->Local_name("name")
                       ->Normalized_value(char_buffer)
                       );
      break;
    case variable:
      char_buffer.insert(char_buffer.begin(),
                         this->basic_object->begin(),
                         this->basic_object->end());
      item->Attributes((new XML_attr_info_item)
                       ->Local_name("name")
                       ->Normalized_value(char_buffer)
                       );
      break;
    case application:
      construct_iterator = this->construct->begin();
      construct_size = this->construct->size();
      construct_end = this->construct->end();
      if (construct_size < 2)
        {
          for (int index = 0; index < 2 - construct_size; ++index)
            this->Application(1,
                              (new OpenMath_object)
                              ->Char_string("[Missing object]")
                              );
          OpenMath_object_p_t om_error;
          delete item;
          item = ((om_error = new OpenMath_object)
                  ->Error(2,
                          this,
                          (new OpenMath_object)
                          ->Char_string("[An OM Application object "
                                        "needs at least 2 subobjects.")
                          )
                  ->To_XML_elem_info_item_subobject()
                  );
          delete om_error;
          break;
        }
      for (int index = 0; index < construct_size; ++index)
        {
          if (construct_iterator == construct_end) break;
          if (*construct_iterator)
            item->Children((*(construct_iterator++))
                           ->To_XML_elem_info_item_subobject());
        }
      break;
    case attribution:
      XML_elem_info_item_p_t atp;
      construct_iterator = this->construct->begin();
      construct_size = this->construct->size();
      construct_end = this->construct->end();
      if (construct_size < 2)
        {
          for (int index = 0; index < 2 - construct_size; ++index)
            this->Attribution(1,
                              (new OpenMath_object)
                              ->Char_string("[Missing object]")
                              );
          OpenMath_object_p_t om_error;
          delete item;
          item = ((om_error = new OpenMath_object)
                  ->Error(2,
                          this,
                          (new OpenMath_object)
                          ->Char_string("[An OM Attribution object "
                                        "needs at least 2 subobjects.")
                          )
                  ->To_XML_elem_info_item_subobject()
                  );
          delete om_error;
          break;
        }
      ((atp = new XML_elem_info_item)
       ->Local_name("OMATP")
       );
      for (int index = 0; index < construct_size - 1; ++index)
        {
          if (construct_iterator == construct_end) break;
          if (*construct_iterator)
            atp->Children((*(construct_iterator++))
                          ->To_XML_elem_info_item_subobject());
        }
      item->Children(atp);
      item->Children((*(construct_iterator++))
                     ->To_XML_elem_info_item_subobject());
      break;
    case binding:
      XML_elem_info_item_p_t bvar;
      construct_iterator = this->construct->begin();
      construct_size = this->construct->size();
      construct_end = this->construct->end();
      if (construct_size < 3)
        {
          for (int index = 0; index < 3 - construct_size; ++index)
            this->Binding(1,
                          (new OpenMath_object)
                          ->Char_string("[Missing object]")
                          );
          OpenMath_object_p_t om_error;
          delete item;
          item = ((om_error = new OpenMath_object)
                  ->Error(2,
                          this,
                          (new OpenMath_object)
                          ->Char_string("[An OM Binding object "
                                        "needs at least 3 subobjects.")
                          )
                  ->To_XML_elem_info_item_subobject()
                  );
          delete om_error;
          break;
        }
      item->Children((*(construct_iterator++))
                     ->To_XML_elem_info_item_subobject());
      ((bvar = new XML_elem_info_item)
       ->Local_name("OMBVAR")
       );
      for (int index = 0; index < construct_size - 2; ++index)
        {
          if (construct_iterator == construct_end) break;
          if (*construct_iterator)
            bvar->Children((*(construct_iterator++))
                           ->To_XML_elem_info_item_subobject());
        }
      item->Children(bvar);
      item->Children((*(construct_iterator++))
                     ->To_XML_elem_info_item_subobject());
      break;
    case error:
      construct_iterator = this->construct->begin();
      construct_size = this->construct->size();
      construct_end = this->construct->end();
      if (construct_size < 2)
        {
          for (int index = 0; index < 2 - construct_size; ++index)
            this->Error(1,
                        (new OpenMath_object)
                        ->Char_string("[Missing object]")
                        );
          OpenMath_object_p_t om_error;
          delete item;
          item = ((om_error = new OpenMath_object)
                  ->Error(2,
                          this,
                          (new OpenMath_object)
                          ->Char_string("[An OM Error object "
                                        "needs at least 2 subobjects.")
                          )
                  ->To_XML_elem_info_item_subobject()
                  );
          delete om_error;
          break;
        }
      for (int index = 0; index < construct_size; ++index)
        {
          if (construct_iterator == construct_end) break;
          if (*construct_iterator)
            item->Children((*(construct_iterator++))
                           ->To_XML_elem_info_item_subobject());
        }
      break;
    }
  
  if (buffer) { free(buffer); buffer = NULL; };
  
  return item;
}

using std::list;

/////////////
// Construct - Return the construct for inspection.
// no arguments.
// returns const list pointer, or NULL if object is not a construct.
//public:
const list<OpenMath_object_p_t>* OpenMath_object::Construct()
{
  return construct;
}

/////////////
// Integer - Set OpenMath integer object value.
// integer: new value for object, encoded as an int.
// returns object pointer for operator concatenation.
//public:
OpenMath_object_p_t OpenMath_object::Integer(int integer)
{
  Type(OpenMath_object::integer);
  basic_object->erase(basic_object->begin(), basic_object->end());
  if (integer == 0)
    {
      basic_object->push_back('0');
      return this;
    }
  char buffer[256];
  sprintf(buffer, "%d", integer);
  
  return Integer(buffer);
}

/////////////
// Integer - Set OpenMath integer object value.
// integer: new value for object, encoded as a char vector.
// returns object pointer for operator concatenation.
//public:
OpenMath_object_p_t OpenMath_object::Integer(vector<char> &integer)
{
  if (!(&integer)) return this;
  
  Type(OpenMath_object::integer);
  basic_object->erase(basic_object->begin(), basic_object->end());
  basic_object->insert(basic_object->begin(), integer.begin(), integer.end());
  
  return this;
}

/////////////
// Integer - Set OpenMath integer object value.
// integer: new value for object, encoded as a C++ char string.
// returns object pointer for operator concatenation.
//public:
OpenMath_object_p_t OpenMath_object::Integer(string &integer)
{
  if (!(&integer)) return this;
  
  Type(OpenMath_object::integer);
  basic_object->erase(basic_object->begin(), basic_object->end());
  basic_object->insert(basic_object->begin(), integer.begin(), integer.end());
  
  return this;
}

/////////////
// Integer - Set OpenMath integer object value.
// integer: new value for object, encoded as a zero-terminated C ASCII string.
// returns object pointer for operator concatenation.
//public:
OpenMath_object_p_t OpenMath_object::Integer(const char* integer)
{
  if (!integer) return this;
  
  Type(OpenMath_object::integer);
  
  return Integer(integer, strlen(integer));
}

/////////////
// Integer - Set OpenMath integer object value.
// integer: new value for object, encoded as a zero-terminated C ASCII string.
// returns object pointer for operator concatenation.
//public:
OpenMath_object_p_t OpenMath_object::Integer(const char* integer, int count)
{
  if (!integer) return this;
  
  Type(OpenMath_object::integer);
  basic_object->erase(basic_object->begin(), basic_object->end());
  for (int i = 0; i < count; ++i)
    basic_object->push_back((unsigned char)integer[i]);
  
  return this;
}

/////////////
// Integer - Access to OpenMath integer object.
// no arguments.
// returns pointer to char vector.
//public:
vector<unsigned char>* OpenMath_object::Integer()
{
  return this->basic_object;
}

/////////////
// Fp_number - Set OpenMath fp_number object value.
// fp_number: new value, encoded as an int.
// returns object pointer for operator concatenation.
//public:
OpenMath_object_p_t OpenMath_object::Fp_number(double fp_number)
{
  Type(OpenMath_object::fp_number);
  this->double_float_number = fp_number;
  
  return this;
}

/////////////
// Fp_number - Set OpenMath fp_number object value.
// fp_number: new value, encoded as a C++ ASCII string.
// returns object pointer for operator concatenation.
//public:
OpenMath_object_p_t OpenMath_object::Fp_number(string &fp_number)
{
  if (!(&fp_number)) return this;
  
  Type(OpenMath_object::fp_number);
  sscanf(fp_number.c_str(), "%lg", &(this->double_float_number));
  
  return this;
}

/////////////
// Fp_number - Set OpenMath fp_number object value.
// fp_number: new value, encoded as a char vector.
// returns object pointer for operator concatenation.
//public:
OpenMath_object_p_t OpenMath_object::Fp_number(vector<char> &fp_number)
{
  if (!(&fp_number)) return this;
  
  Type(OpenMath_object::fp_number);
  
  unsigned int length;
  if (fp_number.size() < SAFE_BUFFER_SIZE_FOR_DOUBLE_FLOAT_TO_STRING)
    length = fp_number.size();
  else
    length = SAFE_BUFFER_SIZE_FOR_DOUBLE_FLOAT_TO_STRING - 1;
  
  char buffer[SAFE_BUFFER_SIZE_FOR_DOUBLE_FLOAT_TO_STRING];
  unsigned int index;
  for (index = 0; index < length; ++index)
    buffer[index] = fp_number[index];
  
  buffer[index] = 0;
  sscanf(buffer, "%lg", &(this->double_float_number));
  
  return this;
}

/////////////
// Fp_number - Set OpenMath fp_number object value.
// fp_number: new value, encoded as a zero-terminated C ASCII string.
// returns object pointer for operator concatenation.
//public:
OpenMath_object_p_t OpenMath_object::Fp_number(const char* fp_number)
{
  if (!fp_number) return this;
  
  Type(OpenMath_object::fp_number);
  sscanf(fp_number, "%lg", &(this->double_float_number));
  
  return this;
}

/////////////
// Fp_number - Set OpenMath fp_number object value.
// fp_number: new value, encoded as a zero-terminated C ASCII string.
// returns object pointer for operator concatenation.
//public:
OpenMath_object_p_t OpenMath_object::Fp_number(const char* fp_number, unsigned int count)
{
  if (!fp_number) return this;
  
  Type(OpenMath_object::fp_number);
  
  unsigned int length;
  if (count < SAFE_BUFFER_SIZE_FOR_DOUBLE_FLOAT_TO_STRING)
    length = count;
  else
    length = SAFE_BUFFER_SIZE_FOR_DOUBLE_FLOAT_TO_STRING - 1;
  
  char buffer[256];
  unsigned int index;
  for (index = 0; index < length; ++index)
    buffer[index] = fp_number[index];
  
  buffer[index] = 0;
  sscanf(buffer, "%lg", &(this->double_float_number));
  
  return this;
}

/////////////
// Fp_number - Access to OpenMath fp_number object.
// no arguments.
// returns double float.
//public:
double OpenMath_object::Fp_number()
{
  return double_float_number;
}

using std::list;

/////////////
// Error - Append objects to an OpenMath error object.
// num_args: Number of objects to append.
// returns object pointer for operator concatenation.
//public:
OpenMath_object_p_t OpenMath_object::Error(int num_args, ...)
{
  Type(OpenMath_object::error);
  if (num_args < 1) return this;
  
  va_list arg_list;
  va_start(arg_list, num_args);
  for (int arg_index = 1; arg_index <= num_args; ++arg_index)
    {
      construct->push_back(va_arg(arg_list, OpenMath_object_p_t));
    }
  va_end(arg_list);
  
  return this;
}


/////////////
// Error - Append objects to an OpenMath error object.
// object_list: const list<OpenMath_object_p_t>* of objects to append.
// returns object pointer for operator concatenation.
//public:
OpenMath_object_p_t OpenMath_object::Error(const list<OpenMath_object_p_t>* object_list)
{
  Type(OpenMath_object::error);
  if (!object_list) return this;
  if (object_list->size() < 1) return this;
  
  list<OpenMath_object_p_t>::const_iterator object_list_iterator;
  for (object_list_iterator = object_list->begin();
       object_list_iterator != object_list->end();
       ++object_list_iterator)
    {
      construct->push_back(*object_list_iterator);
    }
  
  return this;
}

/////////////
// Delete - Delete an object.
// object: OpenMath_object to be destroyed, passed as pointer.
// returns true on success, false on fail.
//public:
//static
bool OpenMath_object::Delete(OpenMath_object_p_t object)
{
  if (!object) return false;
  
  delete object;
  
  return true;
}

/////////////
// Delete_recursive - Delete an object, recursively for constructs.
// object: OpenMath_object to be destroyed, passed as pointer.
// returns true on success, false on fail.
//public:
//static
bool OpenMath_object::Delete_recursive(OpenMath_object_p_t object)
{
  if (!object) return false;
  
  if (object->construct)
    for_each(object->construct->begin(), object->construct->end(),
             &Delete);
  
  delete object;
  
  return true;
}
